from .base import AbstractOrderHandler
from ..event.event import OrderType, OrderEvent
from ..common import get_cur_time
import math


class MaxOrderHandler(AbstractOrderHandler):
    """
    create all-in/all-out order
    TO-DO: Reject the order when cash is not enough 
    """
    def __init__(self, portfolio, event_queue, session_type, db_client=None):
        self.db_client = db_client
        self.portfolio = portfolio
        self.event_queue = event_queue
        self.session_type = session_type
        self.db_client = db_client

    def create_order(self, signal_event):
        fin_instrument = self.portfolio.get_fin_instrument(signal_event.ticker)
        #all-out if it has a position already 
        if fin_instrument is not None:
            quantity = int(abs(fin_instrument.POS))

        #all-in if it does not have any position
        else:   
            if signal_event.order_type == OrderType.MARKET:
                quantity = math.floor(self.portfolio.cash / signal_event.bar_closing_price)
                
            elif signal_event.order_type == OrderType.LIMIT:
                quantity = math.floor(self.portfolio.cash / signal_event.entry_price)
            
            if quantity <= 0:
                print("order rejected due to cash not enough")
                return
            
        cur_time = get_cur_time(signal_event, self.session_type)
        ticker = signal_event.ticker
        order_type = signal_event.order_type
        direction = signal_event.direction
        entry_price = signal_event.entry_price
        order_event = OrderEvent(cur_time, ticker, order_type, 
                                 direction, quantity, entry_price)
        print(order_event)
        self.event_queue.put(order_event)
        
        if self.db_client is not None:
            self._save_to_db(order_event)           
    


from .base import AbstractStrategy
from ..technical_indicators.macd import MACD
from ..technical_indicators.rsi import RSI
from ..event.event import SignalEvent, OrderType, Direction
from ..common import get_cur_time
from enum import Enum
from datetime import datetime, timezone

GO = Enum("GO", "LONG SHORT NONE")  

class MACDRSI(AbstractStrategy):
    #to indicate what kind of direction we want to trade
           
    def __init__(self, ticker_list, macd_shortPeriod, macd_longPeriod,
                 macd_signalPeriod, rsi_period, 
                 event_queue, session_type, portfolio, db_client=None
        ):
        """
        Parameters
        ----------
        tickers_data : dict: {key:string, value: list }
                       key is the ticker symbol, e.g. 'GOOG'
                       list[0] stores MACD object.
                       list[1] stores RSI object.
                       
        session_type: 'backtest' or 'live'
        -------

        """
        self.db_client = db_client
        self.macd_shortPeriod = macd_shortPeriod
        self.macd_longPeriod = macd_longPeriod
        self.macd_signalPeriod = macd_signalPeriod
        self.rsi_period = rsi_period
        self.tickers_data = self._subscribe_ticker(ticker_list)
        self.event_queue = event_queue
        self.session_type = session_type
        self.event_queue = event_queue
        self.session_type = session_type
        self.portfolio = portfolio
        self.go_flag = GO.NONE


    def calculate_signal(self, bar_event):
        ticker_data = self.tickers_data[bar_event.ticker]            
        ticker_data[0].update(bar_event.close_price)
        ticker_data[1].update(bar_event.close_price)
        #Both MACD and RSI have value
        if ticker_data[0] and ticker_data[1] and len(ticker_data[0].data_store[2]) >= 2:                            
            macd = ticker_data[0].lastest_val
            rsi = ticker_data[1].lastest_val
            ticker_info = self.portfolio.get_fin_instrument(bar_event.ticker)
            if self.db_client is not None:
                self.save_to_db_ta(bar_event, macd, rsi)
            #want to enter long/short position
            if ticker_info is None:           
                if self.go_flag != GO.SHORT and self._long_signal(bar_event, macd, rsi, ticker_data):
                    cur_time = get_cur_time(bar_event, self.session_type)
                    signal_event = SignalEvent(cur_time, bar_event.ticker,
                                                OrderType.MARKET, Direction.LONG, 
                                                bar_event.close_price)
                    self.event_queue.put(signal_event)
                    if self.db_client is not None:
                        self.save_to_db_signal(signal_event)                   
                    print(signal_event)
                                            
                if self.go_flag != GO.LONG and self._short_signal(bar_event, macd, rsi, ticker_data):
                    cur_time = get_cur_time(bar_event, self.session_type)                    
                    signal_event = SignalEvent(cur_time, bar_event.ticker,
                                                OrderType.MARKET, Direction.SHORT, 
                                                bar_event.close_price)  
                    self.event_queue.put(signal_event)
                    if self.db_client is not None:
                        self.save_to_db_signal(signal_event)                       
                    print(signal_event)  

            else:                                            
                #want to exit long position                                                  
                if ticker_info.POS > 0 and self._exit_signal_long(bar_event, macd, rsi, ticker_data):                  
                    cur_time = get_cur_time(bar_event, self.session_type)
                    signal_event = SignalEvent(cur_time, bar_event.ticker,
                                            OrderType.MARKET, Direction.SHORT, 
                                            bar_event.close_price)                    
                    self.event_queue.put(signal_event)  
                    if self.db_client is not None:
                        self.save_to_db_signal(signal_event)                       
                    print(signal_event)
                            
                #want to exit short position            
                elif ticker_info.POS < 0 and self._exit_signal_short(bar_event, macd, rsi, ticker_data):              
                    cur_time = get_cur_time(bar_event, self.session_type)
                    signal_event = SignalEvent(cur_time, bar_event.ticker,
                                            OrderType.MARKET, Direction.LONG, 
                                            bar_event.close_price)                                                
                    self.event_queue.put(signal_event) 
                    if self.db_client is not None:
                        self.save_to_db_signal(signal_event)                       
                    print(signal_event)


    def save_to_db_ta(self, bar_event, macd, rsi):
        query = "INSERT INTO MACDRSI VALUES (%s, %s, %s, %s, %s, %s)"
        timestamp_converted = datetime.fromtimestamp(bar_event.timestamp, timezone.utc)
        data = (timestamp_converted, bar_event.ticker, macd[0], macd[1], macd[2], rsi)
        #print(data)
        cursor = self.db_client.cursor()
        cursor.execute(query, data)
        self.db_client.commit()   

                
    def _subscribe_ticker(self, ticker_list):
        tickers_data = {}
        for ticker in ticker_list:
            data = []           
            macd = MACD(self.macd_shortPeriod, self.macd_longPeriod, self.macd_signalPeriod)
            rsi = RSI(self.rsi_period)  
            data = [macd, rsi]
            tickers_data[ticker] = data

        return tickers_data  


    def _long_signal(self, bar_event, macd, rsi, ticker_data):
        rsi_condition = rsi <= 30
        if self.go_flag == GO.NONE and rsi_condition:
            #now we can wait for the macd crosses above signal line
            #print("Want long at:", bar_event.timestamp)
            self.go_flag = GO.LONG
            #print("GO Long Flag set")
            return False

        if self.go_flag == GO.LONG:           
            prev_macd_histogram = ticker_data[0].data_store[2][-2]
            macd_histogram = macd[2]
            #bullish crossover
            if macd_histogram > 0  and prev_macd_histogram <= 0 and rsi > 30: 
                self.go_flag = GO.NONE
                #print("Bullish crossover")
                return True 

            elif rsi >= 70:
                #print("reset long at:", bar_event.timestamp)
                self.go_flag = GO.NONE
                return False              
       
        return False


    def _short_signal(self, bar_event, macd, rsi, ticker_data):
        rsi_condition = rsi >= 70
        if self.go_flag == GO.NONE and rsi_condition:
            #now we can wait for the macd crosses above signal line
            #print("Want short at:", bar_event.timestamp)
            #print("GO Short Flag set")
            self.go_flag = GO.SHORT
            return False

        if self.go_flag == GO.SHORT:
            prev_macd_histogram = ticker_data[0].data_store[2][-2]
            macd_histogram = macd[2]
            #bearish crossover 
            if macd_histogram < 0  and prev_macd_histogram >= 0 and rsi < 70: 
                self.go_flag = GO.NONE
                #print("Bearish corssover")
                return True 

            elif rsi <= 30:
                #print("reset short at:", bar_event.timestamp)
                self.go_flag = GO.NONE
                return False

        return False


    def _exit_signal_long(self, bar_event, macd, rsi, ticker_data):
        macd_histogram = macd[2]
        prev_macd_histogram = ticker_data[0].data_store[2][-2]
        exit_long = macd_histogram < 0  and prev_macd_histogram >= 0
        #print("exit long")
        #if exit_long:
            #print("exit long")
        return exit_long   


    def _exit_signal_short(self, bar_event, macd, rsi, ticker_data):
        macd_histogram = macd[2]
        prev_macd_histogram = ticker_data[0].data_store[2][-2]
        exit_short = macd_histogram > 0  and prev_macd_histogram <= 0
        #if exit_short:
            #print("exit short")        
        #print("exit short")
        return exit_short


    def stop_loss_trigger(self, bar_event):
        pass
        # """
        # stop loss is trigger if the current price is 
        # lower(for long)/higher(for short) than
        # the low of entry candle
        # """
        # invested_type = self.tickers_data[bar_event.ticker][2]
        # if invested_type == InvestedType.LONG and bar_event.close_price < self.stop_price:
        #     print("Long Stop loss:", bar_event.timestamp)
        #     return True

        # if invested_type == InvestedType.SHORT and bar_event.close_price > self.stop_price:
        #     print("Short Stop loss:", bar_event.timestamp)
        #     return True

        # return False      



    def get_ticker(self):
        return self.tickers_data
    
    
    def __str__(self):
        return "MACD-RSI Strategy"
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 20 15:16:52 2021

@author: ron
"""

from .base import AbstractTechnicalIndicator
from .ema import EMA


class MACD(AbstractTechnicalIndicator):
    
    def __init__(self, short_period, long_period, signal_period):
        """
        Parameters
        ----------
        data_store : list of lists
                     First list stores macd value
                     Second list stores macd signal value
                     Third list stores macd histogram value
        """        
        self.ema_short = EMA(short_period)
        self.ema_long = EMA(long_period)
        self.ema_signal = EMA(signal_period)
        self.lastest_val = []
        self.data_store = [[],[],[]]
        self.ready = False
        
        
    def update(self, price):
        self.ema_short.update(price)
        self.ema_long.update(price)
        if self.ema_short and self.ema_long:
            #macd line
            macd = self.ema_short.lastest_val - self.ema_long.lastest_val
            self.ema_signal.update(macd)
            if not self.ready and self.ema_signal:
                self.ready = True
                
            if self.ready:
                macd_signal = self.ema_signal.lastest_val
                macd_histogram = macd - macd_signal
                self.lastest_val = [macd, macd_signal, macd_histogram]
                self.data_store[0].append(macd)
                self.data_store[1].append(macd_signal)
                self.data_store[2].append(macd_histogram)
                               
                        
    def __bool__(self):
        return self.ready
                
                
                
            
            
        
        

            
            
                   
                    
            
                
                
                    
            
            
            
        


"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


from . import comm
from .common import UNSET_DOUBLE
from .object_implem import Object
from .enum_implem import Enum
from .utils import decode

#TODO: add support for Rebate, P/L, ShortableShares conditions 


class OrderCondition(Object):
    Price = 1
    Time = 3
    Margin = 4
    Execution = 5
    Volume = 6
    PercentChange = 7

    def __init__(self, condType):
        self.condType = condType
        self.isConjunctionConnection = True

    def type(self):
        return self.condType

    def And(self):
        self.isConjunctionConnection = True
        return self

    def Or(self):
        self.isConjunctionConnection = False
        return self

    def decode(self, fields):
        connector = decode(str, fields)
        self.isConjunctionConnection = connector == "a"

    def make_fields(self):
        return [comm.make_field("a" if self.isConjunctionConnection else "o")]
 
    def __str__(self):
        return "<AND>" if self.isConjunctionConnection else "<OR>"


class ExecutionCondition(OrderCondition):
    def __init__(self, secType=None, exch=None, symbol=None):
        OrderCondition.__init__(self, OrderCondition.Execution)
        self.secType = secType
        self.exchange = exch
        self.symbol = symbol

    def decode(self, fields):
        OrderCondition.decode(self, fields)
        self.secType = decode(str, fields)
        self.exchange = decode(str, fields)
        self.symbol = decode(str, fields)

    def make_fields(self):
        flds = OrderCondition.make_fields(self) + \
            [comm.make_field(self.secType),
            comm.make_field(self.exchange), 
            comm.make_field(self.symbol)]
        return flds
 
    def __str__(self):
        return "trade occurs for " + self.symbol + " symbol on " + \
            self.exchange + " exchange for " + self.secType + " security type"
        

class OperatorCondition(OrderCondition):
    def __init__(self, condType=None, isMore=None):
        OrderCondition.__init__(self, condType)
        self.isMore = isMore

    def valueToString(self) -> str:
        raise NotImplementedError("abstractmethod!")

    def setValueFromString(self, text: str) -> None:
        raise NotImplementedError("abstractmethod!")

    def decode(self, fields):
        OrderCondition.decode(self, fields)
        self.isMore = decode(bool, fields)
        text = decode(str, fields)
        self.setValueFromString(text)

    def make_fields(self):
        flds = OrderCondition.make_fields(self) + \
            [comm.make_field(self.isMore),
             comm.make_field(self.valueToString()), ]
        return flds

    def __str__(self):
        sb = ">= " if self.isMore else "<= "
        return " %s %s" % (sb, self.valueToString())
 

class MarginCondition(OperatorCondition):
    def __init__(self, isMore=None, percent=None):
        OperatorCondition.__init__(self, OrderCondition.Margin, isMore)
        self.percent = percent

    def decode(self, fields):
        OperatorCondition.decode(self, fields)

    def make_fields(self):
        flds = OperatorCondition.make_fields(self)
        return flds

    def valueToString(self) -> str:
        return str(self.percent)

    def setValueFromString(self, text: str) -> None:
        self.percent = float(text)
 
    def __str__(self):
        return "the margin cushion percent %s " % (
            OperatorCondition.__str__(self))
 

class ContractCondition(OperatorCondition):
    def __init__(self, condType=None, conId=None, exch=None, isMore=None):
        OperatorCondition.__init__(self, condType, isMore)
        self.conId = conId
        self.exchange = exch

    def decode(self, fields):
        OperatorCondition.decode(self, fields)
        self.conId = decode(int, fields)
        self.exchange = decode(str, fields)

    def make_fields(self):
        flds = OperatorCondition.make_fields(self) + \
            [comm.make_field(self.conId),
             comm.make_field(self.exchange), ]
        return flds

    def __str__(self):
        return "%s on %s is %s " % (self.conId, self.exchange, 
            OperatorCondition.__str__(self))
 

class TimeCondition(OperatorCondition):
    def __init__(self, isMore=None, time=None):
        OperatorCondition.__init__(self, OrderCondition.Time, isMore)
        self.time = time

    def decode(self, fields):
        OperatorCondition.decode(self, fields)

    def make_fields(self):
        flds = OperatorCondition.make_fields(self)
        return flds

    def valueToString(self) -> str:
        return self.time

    def setValueFromString(self, text: str) -> None:
        self.time = text
 
    def __str__(self):
        return "time is %s " % (OperatorCondition.__str__(self))  

                          
class PriceCondition(ContractCondition):
    TriggerMethodEnum = Enum(
          "Default", # = 0,
          "DoubleBidAsk", # = 1,
          "Last", # = 2,
          "DoubleLast", # = 3,
          "BidAsk", # = 4,
          "N/A1",
          "N/A2",
          "LastBidAsk", #= 7,
          "MidPoint") #= 8
     
    def __init__(self, triggerMethod=None, conId=None, exch=None, isMore=None, 
                 price=None):
        ContractCondition.__init__(self, OrderCondition.Price, conId, exch, 
                                   isMore)
        self.price = price
        self.triggerMethod = triggerMethod

    def decode(self, fields):
        ContractCondition.decode(self, fields)
        self.triggerMethod = decode(int, fields)

    def make_fields(self):
        flds = ContractCondition.make_fields(self) + \
            [comm.make_field(self.triggerMethod), ]
        return flds

    def valueToString(self) -> str:
        return str(self.price)

    def setValueFromString(self, text: str) -> None:
        self.price = float(text)
 
    def __str__(self):
        return "%s price of %s " % (
            PriceCondition.TriggerMethodEnum.to_str(self.triggerMethod), 
            ContractCondition.__str__(self))


class PercentChangeCondition(ContractCondition):
    def __init__(self, conId=None, exch=None, isMore=None,
                 changePercent=UNSET_DOUBLE):
        ContractCondition.__init__(self, OrderCondition.PercentChange, conId,
                                   exch, isMore)
        self.changePercent = changePercent

    def decode(self, fields):
        ContractCondition.decode(self, fields)

    def make_fields(self):
        flds = ContractCondition.make_fields(self)
        return flds

    def valueToString(self) -> str:
        return str(self.changePercent)

    def setValueFromString(self, text: str) -> None:
        self.changePercent = float(text)

    def __str__(self):
        return "percent change of %s " % (
            ContractCondition.__str__(self))
  

class VolumeCondition(ContractCondition):
    def __init__(self, conId=None, exch=None, isMore=None, volume=None):
        ContractCondition.__init__(self, OrderCondition.Volume, conId, exch, 
                                   isMore)
        self.volume = volume

    def decode(self, fields):
        ContractCondition.decode(self, fields)

    def make_fields(self):
        flds = ContractCondition.make_fields(self)
        return flds

    def valueToString(self) -> str:
        return str(self.volume)

    def setValueFromString(self, text: str) -> None:
        self.volume = int(text)

    def __str__(self):
        return "volume of %s " % (
            ContractCondition.__str__(self))
 

def Create(condType):
    cond = None

    if OrderCondition.Execution == condType:
        cond = ExecutionCondition()
    elif OrderCondition.Margin == condType:
        cond = MarginCondition()
    elif OrderCondition.PercentChange == condType:
        cond = PercentChangeCondition()
    elif OrderCondition.Price == condType:
        cond = PriceCondition()
    elif OrderCondition.Time == condType:
        cond = TimeCondition()
    elif OrderCondition.Volume == condType:
        cond = VolumeCondition()

    return cond



"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


"""
High level IB message info.
"""


# field types
INT = 1
STR = 2
FLT = 3


# incoming msg id's
class IN:
    TICK_PRICE                = 1
    TICK_SIZE                 = 2
    ORDER_STATUS              = 3
    ERR_MSG                   = 4
    OPEN_ORDER                = 5
    ACCT_VALUE                = 6
    PORTFOLIO_VALUE           = 7
    ACCT_UPDATE_TIME          = 8
    NEXT_VALID_ID             = 9
    CONTRACT_DATA             = 10
    EXECUTION_DATA            = 11
    MARKET_DEPTH              = 12
    MARKET_DEPTH_L2           = 13
    NEWS_BULLETINS            = 14
    MANAGED_ACCTS             = 15
    RECEIVE_FA                = 16
    HISTORICAL_DATA           = 17
    BOND_CONTRACT_DATA        = 18
    SCANNER_PARAMETERS        = 19
    SCANNER_DATA              = 20
    TICK_OPTION_COMPUTATION   = 21
    TICK_GENERIC              = 45
    TICK_STRING               = 46
    TICK_EFP                  = 47
    CURRENT_TIME              = 49
    REAL_TIME_BARS            = 50
    FUNDAMENTAL_DATA          = 51
    CONTRACT_DATA_END         = 52
    OPEN_ORDER_END            = 53
    ACCT_DOWNLOAD_END         = 54
    EXECUTION_DATA_END        = 55
    DELTA_NEUTRAL_VALIDATION  = 56
    TICK_SNAPSHOT_END         = 57
    MARKET_DATA_TYPE          = 58
    COMMISSION_REPORT         = 59
    POSITION_DATA             = 61
    POSITION_END              = 62
    ACCOUNT_SUMMARY           = 63
    ACCOUNT_SUMMARY_END       = 64
    VERIFY_MESSAGE_API        = 65
    VERIFY_COMPLETED          = 66
    DISPLAY_GROUP_LIST        = 67
    DISPLAY_GROUP_UPDATED     = 68
    VERIFY_AND_AUTH_MESSAGE_API = 69
    VERIFY_AND_AUTH_COMPLETED   = 70
    POSITION_MULTI            = 71
    POSITION_MULTI_END        = 72
    ACCOUNT_UPDATE_MULTI      = 73
    ACCOUNT_UPDATE_MULTI_END  = 74
    SECURITY_DEFINITION_OPTION_PARAMETER = 75
    SECURITY_DEFINITION_OPTION_PARAMETER_END = 76
    SOFT_DOLLAR_TIERS         = 77
    FAMILY_CODES              = 78
    SYMBOL_SAMPLES            = 79
    MKT_DEPTH_EXCHANGES       = 80
    TICK_REQ_PARAMS           = 81
    SMART_COMPONENTS          = 82
    NEWS_ARTICLE              = 83
    TICK_NEWS                 = 84
    NEWS_PROVIDERS            = 85
    HISTORICAL_NEWS           = 86
    HISTORICAL_NEWS_END       = 87
    HEAD_TIMESTAMP            = 88
    HISTOGRAM_DATA            = 89
    HISTORICAL_DATA_UPDATE    = 90
    REROUTE_MKT_DATA_REQ      = 91
    REROUTE_MKT_DEPTH_REQ     = 92
    MARKET_RULE               = 93
    PNL                       = 94
    PNL_SINGLE                = 95
    HISTORICAL_TICKS          = 96
    HISTORICAL_TICKS_BID_ASK  = 97
    HISTORICAL_TICKS_LAST     = 98
    TICK_BY_TICK              = 99
    ORDER_BOUND               = 100
    COMPLETED_ORDER           = 101
    COMPLETED_ORDERS_END      = 102
    REPLACE_FA_END            = 103
    WSH_META_DATA             = 104
    WSH_EVENT_DATA            = 105

# outgoing msg id's
class OUT:
    REQ_MKT_DATA                  = 1
    CANCEL_MKT_DATA               = 2
    PLACE_ORDER                   = 3
    CANCEL_ORDER                  = 4
    REQ_OPEN_ORDERS               = 5
    REQ_ACCT_DATA                 = 6
    REQ_EXECUTIONS                = 7
    REQ_IDS                       = 8
    REQ_CONTRACT_DATA             = 9
    REQ_MKT_DEPTH                 = 10
    CANCEL_MKT_DEPTH              = 11
    REQ_NEWS_BULLETINS            = 12
    CANCEL_NEWS_BULLETINS         = 13
    SET_SERVER_LOGLEVEL           = 14
    REQ_AUTO_OPEN_ORDERS          = 15
    REQ_ALL_OPEN_ORDERS           = 16
    REQ_MANAGED_ACCTS             = 17
    REQ_FA                        = 18
    REPLACE_FA                    = 19
    REQ_HISTORICAL_DATA           = 20
    EXERCISE_OPTIONS              = 21
    REQ_SCANNER_SUBSCRIPTION      = 22
    CANCEL_SCANNER_SUBSCRIPTION   = 23
    REQ_SCANNER_PARAMETERS        = 24
    CANCEL_HISTORICAL_DATA        = 25
    REQ_CURRENT_TIME              = 49
    REQ_REAL_TIME_BARS            = 50
    CANCEL_REAL_TIME_BARS         = 51
    REQ_FUNDAMENTAL_DATA          = 52
    CANCEL_FUNDAMENTAL_DATA       = 53
    REQ_CALC_IMPLIED_VOLAT        = 54
    REQ_CALC_OPTION_PRICE         = 55
    CANCEL_CALC_IMPLIED_VOLAT     = 56
    CANCEL_CALC_OPTION_PRICE      = 57
    REQ_GLOBAL_CANCEL             = 58
    REQ_MARKET_DATA_TYPE          = 59
    REQ_POSITIONS                 = 61
    REQ_ACCOUNT_SUMMARY           = 62
    CANCEL_ACCOUNT_SUMMARY        = 63
    CANCEL_POSITIONS              = 64
    VERIFY_REQUEST                = 65
    VERIFY_MESSAGE                = 66
    QUERY_DISPLAY_GROUPS          = 67
    SUBSCRIBE_TO_GROUP_EVENTS     = 68
    UPDATE_DISPLAY_GROUP          = 69
    UNSUBSCRIBE_FROM_GROUP_EVENTS = 70
    START_API                     = 71
    VERIFY_AND_AUTH_REQUEST       = 72
    VERIFY_AND_AUTH_MESSAGE       = 73
    REQ_POSITIONS_MULTI           = 74
    CANCEL_POSITIONS_MULTI        = 75
    REQ_ACCOUNT_UPDATES_MULTI     = 76
    CANCEL_ACCOUNT_UPDATES_MULTI  = 77
    REQ_SEC_DEF_OPT_PARAMS        = 78
    REQ_SOFT_DOLLAR_TIERS         = 79
    REQ_FAMILY_CODES              = 80
    REQ_MATCHING_SYMBOLS          = 81
    REQ_MKT_DEPTH_EXCHANGES       = 82
    REQ_SMART_COMPONENTS          = 83
    REQ_NEWS_ARTICLE              = 84
    REQ_NEWS_PROVIDERS            = 85
    REQ_HISTORICAL_NEWS           = 86
    REQ_HEAD_TIMESTAMP            = 87
    REQ_HISTOGRAM_DATA            = 88
    CANCEL_HISTOGRAM_DATA         = 89
    CANCEL_HEAD_TIMESTAMP         = 90
    REQ_MARKET_RULE               = 91
    REQ_PNL                       = 92
    CANCEL_PNL                    = 93
    REQ_PNL_SINGLE                = 94
    CANCEL_PNL_SINGLE             = 95
    REQ_HISTORICAL_TICKS          = 96
    REQ_TICK_BY_TICK_DATA         = 97
    CANCEL_TICK_BY_TICK_DATA      = 98
    REQ_COMPLETED_ORDERS          = 99
    REQ_WSH_META_DATA             = 100
    CANCEL_WSH_META_DATA          = 101
    REQ_WSH_EVENT_DATA            = 102
    CANCEL_WSH_EVENT_DATA         = 103



"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


""" 
    Simple enum implementation
"""


class Enum:
    def __init__(self, *args):
        self.idx2name = {}
        for (idx, name) in enumerate(args):
            setattr(self, name, idx)
            self.idx2name[idx] = name

    def to_str(self, idx):
        return self.idx2name.get(idx, "NOTFOUND")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec 29 21:09:43 2021

@author: ron
"""

from datetime import datetime, timezone
import time
from enum import Enum
from tracemalloc import start
import os
import subprocess

SessionType = Enum("SessionType", "LIVE BACKTEST")

def get_cur_time(event, session_type):
    if session_type == SessionType.BACKTEST:
        return event.timestamp
    
    if session_type ==  SessionType.LIVE:
        #return datetime.now(timezone.utc)
        return int(time.time())


class RealTimeBarError(Exception):
    pass


class Bar():
    def __init__(
        self, start_time, end_time,
        open_price, high_price, low_price,
        close_price, volume, wap, count
        ):
        """
        Initialises the BarEvent.
        Parameters:
        timestamp - The timestamp of the Bar event generated (i.e. timestamp of a bar)
        ticker - The ticker symbol, e.g. 'GOOG'.
        open_price - The opening price of the bar
        high_price - The high price of the bar
        low_price - The low price of the bar
        close_price - The close price of the bar
        volume - The volume of trading within the bar
        """
        self.start_time = start_time
        self.end_time = end_time
        self.open_price = open_price
        self.high_price = high_price
        self.low_price = low_price
        self.close_price = close_price
        self.volume = volume
        self.wap = wap
        self.count = count

    def __str__(self):
        format_str = "Start: %s, End: %s, " \
            "Open: %s, High: %s, Low: %s, Close: %s, Volume: %s ,WAP %s, Count %s" % (
                str(self.start_time), str(self.end_time),
                str(self.open_price), str(self.high_price), 
                str(self.low_price), str(self.close_price),
                str(self.volume), str(self.wap), str(self.count)
            )
        return format_str

    def __repr__(self):
        return str(self)


class Error():
    def __init__(self, time, errorCode, errorString):
        self.time = time
        self.code = errorCode
        self.string = errorString


def check_ping():
    proc = subprocess.Popen(
        ['ping', '-q', '-c', '3', "8.8.8.8"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL) 
    proc.wait()
    # and then check the response...
    if proc.returncode == 0:
        return True
    else:
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 26 20:06:59 2021

@author: ron
"""
import unittest
from tradingsystem.statistics.statistics import Statistics
from datetime import datetime, timedelta

class TestStatistics(unittest.TestCase):
    """
    Testing if drawdown and reutrn are correct
    """        
    def test_update(self):
        stat = Statistics(10000)
        date = datetime(2021, 1, 1)
        stat.update(date, 10000)

        date += timedelta(days=1)
        stat.update(date, 10000)

        date += timedelta(days=1)
        stat.update(date, 12000)

        date += timedelta(days=1)
        stat.update(date, 11000)

        date += timedelta(days=1)
        stat.update(date, 15000)

        date += timedelta(days=1)
        stat.update(date, 9000)

        date += timedelta(days=1)
        stat.update(date, 9000)

        date += timedelta(days=1)
        stat.update(date, 10500)

        result = stat.get_results()
        self.assertEqual(result["return"], 5)
        self.assertEqual(result["max_drawdown"], -0.4)
        self.assertEqual(round(result["sharpe_ratio"], 3), 2.29)          



if __name__ == '__main__':
    unittest.main(verbosity=2)


from .base import AbstractDataHandler
import yfinance as yf
import pandas as pd
from ..event.event import BarEvent


class HistoricalBarHandler(AbstractDataHandler):
    def __init__(self, tickers_data, event_queue, db_client=None):
        """
        Parameters
        ----------
        tickers_data : dict{key: ticker, value: dataframe}
        The dataframe must have the foramt below:
        Datetime object, Open, High, Low, Close, Volume
        """ 
        self.db_client = db_client
        self.bar_data = self._sort_tickers_data(tickers_data)
        self.event_queue = event_queue
        self.tickers_lastest_bar = {}
        
           
    def get_next(self):
        try:
            current_bar = next(self.bar_data)
            bar_event = self._create_bar_event(current_bar)
            self.tickers_lastest_bar[bar_event.ticker] = bar_event.close_price 
            self.event_queue.put(bar_event)
            print(bar_event)
            
        except StopIteration:
            raise StopIteration

            
    def get_last(self, ticker):
         """
         Returns the most recent closing price.
         """        
         if ticker in self.tickers_lastest_bar:
             return self.tickers_lastest_bar[ticker]  
         else:
            print("ticker doesn't exist")
            return None

                              
    def _sort_tickers_data(self, tickers_data):
        """
        sort tickers_data according to the date and
        store them in self.tickers_data
        """
        sorted_bar_data = pd.DataFrame()
        #add 'Ticker' column to respective data and then put it in
        #sorted_bar_data
        for ticker in tickers_data:
            tickers_data[ticker]['Ticker'] = ticker
            sorted_bar_data = pd.concat([sorted_bar_data, tickers_data[ticker]])           
        sorted_bar_data.sort_index(inplace=True)
        return sorted_bar_data.itertuples()
    
    
    def _create_bar_event(self, current_bar):
        timestamp = current_bar[0]
        open_price = current_bar[1]
        high_price = current_bar[2]
        low_price = current_bar[3]
        close_price = current_bar[4]
        volume = current_bar[5]
        ticker = current_bar[6]
        return BarEvent(timestamp, ticker, open_price, high_price,
                 low_price, close_price, volume)
    
        
    
           
def get_data_from_yahoo_finance(ticker, start_date, end_date, interval):  
    historical_bar_data = yf.download(ticker, start=start_date, 
                                  end=end_date, interval=interval)
    historical_bar_data.drop('Adj Close', inplace=True, axis=1)
    return historical_bar_data

    
def get_data_from_csv(csv_path):
    historical_bar_data = pd.read_csv(csv_path, index_col=0, parse_dates=True)
    return  historical_bar_data
    
    
    

    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 22 20:12:56 2021

@author: ron
"""

import unittest
from tradingsystem.technical_indicators.sma import SMA
from tradingsystem.technical_indicators.bb import BB
from tradingsystem.technical_indicators.rsi import RSI
from tradingsystem.technical_indicators.ema import EMA
from tradingsystem.technical_indicators.macd import MACD
import talib as ta
import math
import yfinance as yf

class TestIndicators(unittest.TestCase):
    #test indicators and compare the result with talib 
    def setUp(self):
        self.historical_bar_data = yf.download('TSLA', start='2021-01-01', end='2021-11-01')['Close']

      
    def test_sma(self):
        sma = SMA(20)
        tasma = ta.SMA(self.historical_bar_data, 20)
        tasma = [round(val,3) for val in tasma if not math.isnan(val)]
        for val in self.historical_bar_data:
            sma.update(val)
        sma_round = [round(val,3) for val in sma.data_store]
        self.assertListEqual(sma_round, tasma)

        
    def test_bb(self):
        bb = BB(20, 2)
        tabb = ta.BBANDS(self.historical_bar_data, timeperiod=20)
        tabb = [[round(val,3) for val in band if not math.isnan(val)] for band in tabb]
        for val in self.historical_bar_data:
            bb.update(val)            
        bb_round = [[round(val,3) for val in band] for band in bb.data_store]
        self.assertListEqual(bb_round, tabb)

        
    def test_rsi(self):
        rsi = RSI(20)
        tarsi = ta.RSI(self.historical_bar_data, 20)
        tarsi = [round(val,3) for val in tarsi if not math.isnan(val)]
        for val in self.historical_bar_data:
            rsi.update(val)
        rsi_round = [round(val,3) for val in rsi.data_store]
        self.assertListEqual(rsi_round, tarsi)    
        
    
    def test_ema(self):
        ema = EMA(14)
        taema = ta.EMA(self.historical_bar_data, 14)
        taema = [round(val,3) for val in taema if not math.isnan(val)]
        for val in self.historical_bar_data:
            ema.update(val)
        ema_round = [round(val,3) for val in ema.data_store]
        self.assertListEqual(ema_round, taema) 


    def test_macd(self):
        """
        Not test with talib since the MACD formula in talib is different
        """
        self.data = [459.99, 448.85, 446.06, 450.81, 442.8, 448.97, 444.57, 441.4,
                     430.47, 420.05, 431.14, 425.66, 430.58, 431.72, 437.87, 428.43,
                     428.35, 432.5, 443.66, 455.72, 454.49, 452.08, 452.73, 461.91, 
                     463.58, 461.14, 452.08, 442.66, 428.91, 429.79, 431.99, 427.72,
                     423.2, 426.21, 426.98, 435.69]     
        
        macd = MACD(12,26,9)
        expected_result = [[-2.071, -2.622, -2.329], [3.038, 1.906, 1.059], [-5.108, -4.527, -3.388]]
        for val in self.data:
            macd.update(val)
        macd_round = [[round(val,3) for val in r] for r in macd.data_store]
        self.assertListEqual(macd_round, expected_result)             
        
        

if __name__ == '__main__':
    unittest.main(verbosity=2)


"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

"""
The Decoder knows how to transform a message's payload into higher level
IB message (eg: order info, mkt data, etc).
It will call the corresponding method from the EWrapper so that customer's code
(eg: class derived from EWrapper) can make further use of the data.
"""

from .message import IN
from .wrapper import * # @UnusedWildImport
from .contract import ContractDescription
from .server_versions import * # @UnusedWildImport
from .utils import * # @UnusedWildImport
from .softdollartier import SoftDollarTier
from .ticktype import * # @UnusedWildImport
from .tag_value import TagValue
from .scanner import ScanData
from .errors import BAD_MESSAGE
from .common import * # @UnusedWildImport
from .orderdecoder import OrderDecoder

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)
# ch = logging.StreamHandler()
# ch.setLevel(level=logging.DEBUG)
# logger.addHandler(ch)


class HandleInfo(Object):
    def __init__(self, wrap=None, proc=None):
        self.wrapperMeth = wrap
        self.wrapperParams = None
        self.processMeth = proc
        if wrap is None and proc is None:
            raise ValueError("both wrap and proc can't be None")

    def __str__(self):
        s = "wrap:%s meth:%s prms:%s" % (self.wrapperMeth,
                self.processMeth, self.wrapperParams)
        return s


class Decoder(Object):
    def __init__(self, wrapper, serverVersion):
        self.wrapper = wrapper
        self.serverVersion = serverVersion
        self.discoverParams()


    def processTickPriceMsg(self, fields):
        next(fields)
        decode(int, fields)

        reqId = decode(int, fields)
        tickType = decode(int, fields)
        price = decode(float, fields)
        size = decode(Decimal, fields) # ver 2 field
        attrMask = decode(int, fields) # ver 3 field

        attrib = TickAttrib()

        attrib.canAutoExecute = attrMask == 1

        if self.serverVersion >= MIN_SERVER_VER_PAST_LIMIT:
            attrib.canAutoExecute = attrMask & 1 != 0
            attrib.pastLimit = attrMask & 2 != 0
            if self.serverVersion >= MIN_SERVER_VER_PRE_OPEN_BID_ASK:
                attrib.preOpen = attrMask & 4 != 0

        self.wrapper.tickPrice(reqId, tickType, price, attrib)

        # process ver 2 fields
        sizeTickType = TickTypeEnum.NOT_SET
        if TickTypeEnum.BID == tickType:
            sizeTickType = TickTypeEnum.BID_SIZE
        elif TickTypeEnum.ASK == tickType:
            sizeTickType = TickTypeEnum.ASK_SIZE
        elif TickTypeEnum.LAST == tickType:
            sizeTickType = TickTypeEnum.LAST_SIZE
        elif TickTypeEnum.DELAYED_BID == tickType:
            sizeTickType = TickTypeEnum.DELAYED_BID_SIZE
        elif TickTypeEnum.DELAYED_ASK == tickType:
            sizeTickType = TickTypeEnum.DELAYED_ASK_SIZE
        elif TickTypeEnum.DELAYED_LAST == tickType:
            sizeTickType = TickTypeEnum.DELAYED_LAST_SIZE

        if sizeTickType != TickTypeEnum.NOT_SET:
            self.wrapper.tickSize(reqId, sizeTickType, size)


    def processTickSizeMsg(self, fields):
        next(fields)
        decode(int, fields)

        reqId = decode(int, fields)
        sizeTickType = decode(int, fields)
        size = decode(Decimal, fields)

        if sizeTickType != TickTypeEnum.NOT_SET:
            self.wrapper.tickSize(reqId, sizeTickType, size)

    def processOrderStatusMsg(self, fields):
        next(fields)
        if self.serverVersion < MIN_SERVER_VER_MARKET_CAP_PRICE:
            decode(int, fields)
        orderId = decode(int, fields)
        status = decode(str, fields)
        filled = decode(Decimal, fields)
        remaining = decode(Decimal, fields)
        avgFillPrice = decode(float, fields)

        permId = decode(int, fields) # ver 2 field
        parentId = decode(int, fields) # ver 3 field
        lastFillPrice = decode(float, fields) # ver 4 field
        clientId = decode(int, fields) # ver 5 field
        whyHeld = decode(str, fields) # ver 6 field

        if self.serverVersion >= MIN_SERVER_VER_MARKET_CAP_PRICE:
            mktCapPrice = decode(float, fields)
        else:
            mktCapPrice = None

        self.wrapper.orderStatus(orderId, status, filled, remaining,
            avgFillPrice, permId, parentId, lastFillPrice, clientId, whyHeld, mktCapPrice)


    def processOpenOrder(self, fields):
        next(fields)
        
        order = Order()
        contract = Contract()
        orderState = OrderState()

        if self.serverVersion < MIN_SERVER_VER_ORDER_CONTAINER:
            version = decode(int, fields)
        else:
            version = self.serverVersion

        
        OrderDecoder.__init__(self, contract, order, orderState, version, self.serverVersion)

        # read orderId
        OrderDecoder.decodeOrderId(self, fields)

        # read contract fields
        OrderDecoder.decodeContractFields(self, fields)

        # read order fields
        OrderDecoder.decodeAction(self, fields)
        OrderDecoder.decodeTotalQuantity(self, fields)
        OrderDecoder.decodeOrderType(self, fields)
        OrderDecoder.decodeLmtPrice(self, fields)
        OrderDecoder.decodeAuxPrice(self, fields)
        OrderDecoder.decodeTIF(self, fields)
        OrderDecoder.decodeOcaGroup(self, fields)
        OrderDecoder.decodeAccount(self, fields)
        OrderDecoder.decodeOpenClose(self, fields)
        OrderDecoder.decodeOrigin(self, fields)
        OrderDecoder.decodeOrderRef(self, fields)
        OrderDecoder.decodeClientId(self, fields)
        OrderDecoder.decodePermId(self, fields)
        OrderDecoder.decodeOutsideRth(self, fields)
        OrderDecoder.decodeHidden(self, fields)
        OrderDecoder.decodeDiscretionaryAmt(self, fields)
        OrderDecoder.decodeGoodAfterTime(self, fields)
        OrderDecoder.skipSharesAllocation(self, fields)
        OrderDecoder.decodeFAParams(self, fields)
        OrderDecoder.decodeModelCode(self, fields)
        OrderDecoder.decodeGoodTillDate(self, fields)
        OrderDecoder.decodeRule80A(self, fields)
        OrderDecoder.decodePercentOffset(self, fields)
        OrderDecoder.decodeSettlingFirm(self, fields)
        OrderDecoder.decodeShortSaleParams(self, fields)
        OrderDecoder.decodeAuctionStrategy(self, fields)
        OrderDecoder.decodeBoxOrderParams(self, fields)
        OrderDecoder.decodePegToStkOrVolOrderParams(self, fields)
        OrderDecoder.decodeDisplaySize(self, fields)
        OrderDecoder.decodeBlockOrder(self, fields)
        OrderDecoder.decodeSweepToFill(self, fields)
        OrderDecoder.decodeAllOrNone(self, fields)
        OrderDecoder.decodeMinQty(self, fields)
        OrderDecoder.decodeOcaType(self, fields)
        OrderDecoder.skipETradeOnly(self, fields)
        OrderDecoder.skipFirmQuoteOnly(self, fields)
        OrderDecoder.skipNbboPriceCap(self, fields)
        OrderDecoder.decodeParentId(self, fields)
        OrderDecoder.decodeTriggerMethod(self, fields)
        OrderDecoder.decodeVolOrderParams(self, fields, True)
        OrderDecoder.decodeTrailParams(self, fields)
        OrderDecoder.decodeBasisPoints(self, fields)
        OrderDecoder.decodeComboLegs(self, fields)
        OrderDecoder.decodeSmartComboRoutingParams(self, fields)
        OrderDecoder.decodeScaleOrderParams(self, fields)
        OrderDecoder.decodeHedgeParams(self, fields)
        OrderDecoder.decodeOptOutSmartRouting(self, fields)
        OrderDecoder.decodeClearingParams(self, fields)
        OrderDecoder.decodeNotHeld(self, fields)
        OrderDecoder.decodeDeltaNeutral(self, fields)
        OrderDecoder.decodeAlgoParams(self, fields)
        OrderDecoder.decodeSolicited(self, fields)
        OrderDecoder.decodeWhatIfInfoAndCommission(self, fields)
        OrderDecoder.decodeVolRandomizeFlags(self, fields)
        OrderDecoder.decodePegToBenchParams(self, fields)
        OrderDecoder.decodeConditions(self, fields)
        OrderDecoder.decodeAdjustedOrderParams(self, fields)
        OrderDecoder.decodeSoftDollarTier(self, fields)
        OrderDecoder.decodeCashQty(self, fields)
        OrderDecoder.decodeDontUseAutoPriceForHedge(self, fields)
        OrderDecoder.decodeIsOmsContainers(self, fields)
        OrderDecoder.decodeDiscretionaryUpToLimitPrice(self, fields)
        OrderDecoder.decodeUsePriceMgmtAlgo(self, fields)
        OrderDecoder.decodeDuration(self, fields)
        OrderDecoder.decodePostToAts(self, fields)
        OrderDecoder.decodeAutoCancelParent(self, fields, MIN_SERVER_VER_AUTO_CANCEL_PARENT)

        self.wrapper.openOrder(order.orderId, contract, order, orderState)


    def processPortfolioValueMsg(self, fields):

        next(fields)
        version = decode(int, fields)

        # read contract fields
        contract = Contract()
        contract.conId = decode(int, fields) # ver 6 field
        contract.symbol = decode(str, fields)
        contract.secType = decode(str, fields)
        contract.lastTradeDateOrContractMonth = decode(str, fields)
        contract.strike = decode(float, fields)
        contract.right = decode(str, fields)

        if version >= 7:
            contract.multiplier = decode(str, fields)
            contract.primaryExchange = decode(str, fields)

        contract.currency = decode(str, fields)
        contract.localSymbol = decode(str, fields) # ver 2 field
        if version >= 8:
            contract.tradingClass = decode(str, fields)

        position = decode(Decimal, fields)

        marketPrice = decode(float, fields)
        marketValue = decode(float, fields)
        averageCost = decode(float, fields) # ver 3 field
        unrealizedPNL = decode(float, fields) # ver 3 field
        realizedPNL = decode(float, fields) # ver 3 field

        accountName = decode(str, fields) # ver 4 field

        if version == 6 and self.serverVersion == 39:
            contract.primaryExchange = decode(str, fields)

        self.wrapper.updatePortfolio( contract,
            position, marketPrice, marketValue, averageCost,
            unrealizedPNL, realizedPNL, accountName)


    def processContractDataMsg(self, fields):

        next(fields)
        version = 8
        if self.serverVersion < MIN_SERVER_VER_SIZE_RULES:
            version = decode(int, fields)

        reqId = -1
        if version >= 3:
            reqId = decode(int, fields)

        contract = ContractDetails()
        contract.contract.symbol = decode(str, fields)
        contract.contract.secType = decode(str, fields)
        self.readLastTradeDate(fields, contract, False)
        contract.contract.strike = decode(float, fields)
        contract.contract.right = decode(str, fields)
        contract.contract.exchange = decode(str, fields)
        contract.contract.currency = decode(str, fields)
        contract.contract.localSymbol = decode(str, fields)
        contract.marketName = decode(str, fields)
        contract.contract.tradingClass = decode(str, fields)
        contract.contract.conId = decode(int, fields)
        contract.minTick = decode(float, fields)
        if self.serverVersion >= MIN_SERVER_VER_MD_SIZE_MULTIPLIER and self.serverVersion < MIN_SERVER_VER_SIZE_RULES:
            decode(int, fields) # mdSizeMultiplier - not used anymore
        contract.contract.multiplier = decode(str, fields)
        contract.orderTypes = decode(str, fields)
        contract.validExchanges = decode(str, fields)
        contract.priceMagnifier = decode(int, fields) # ver 2 field
        if version >= 4:
            contract.underConId = decode(int, fields)
        if version >= 5:
            contract.longName = decode(str, fields).encode().decode('unicode-escape') if self.serverVersion >= MIN_SERVER_VER_ENCODE_MSG_ASCII7 else decode(str, fields)
            contract.contract.primaryExchange = decode(str, fields)
        if version >= 6:
            contract.contractMonth = decode(str, fields)
            contract.industry = decode(str, fields)
            contract.category = decode(str, fields)
            contract.subcategory = decode(str, fields)
            contract.timeZoneId = decode(str, fields)
            contract.tradingHours = decode(str, fields)
            contract.liquidHours = decode(str, fields)
        if version >= 8:
            contract.evRule = decode(str, fields)
            contract.evMultiplier = decode(int, fields)
        if version >= 7:
            secIdListCount = decode(int, fields)
            if secIdListCount > 0:
                contract.secIdList = []
                for _ in range(secIdListCount):
                    tagValue = TagValue()
                    tagValue.tag = decode(str, fields)
                    tagValue.value = decode(str, fields)
                    contract.secIdList.append(tagValue)

        if self.serverVersion >= MIN_SERVER_VER_AGG_GROUP:
            contract.aggGroup = decode(int, fields)

        if self.serverVersion >= MIN_SERVER_VER_UNDERLYING_INFO:
            contract.underSymbol = decode(str, fields)
            contract.underSecType = decode(str, fields)

        if self.serverVersion >= MIN_SERVER_VER_MARKET_RULES:
            contract.marketRuleIds = decode(str, fields)

        if self.serverVersion >= MIN_SERVER_VER_REAL_EXPIRATION_DATE:
            contract.realExpirationDate = decode(str, fields)

        if self.serverVersion >= MIN_SERVER_VER_STOCK_TYPE:
            contract.stockType = decode(str, fields)

        if self.serverVersion >= MIN_SERVER_VER_FRACTIONAL_SIZE_SUPPORT and self.serverVersion < MIN_SERVER_VER_SIZE_RULES:
            decode(Decimal, fields) # sizeMinTick - not used anymore

        if self.serverVersion >= MIN_SERVER_VER_SIZE_RULES:
            contract.minSize = decode(Decimal, fields)
            contract.sizeIncrement = decode(Decimal, fields)
            contract.suggestedSizeIncrement = decode(Decimal, fields)

        self.wrapper.contractDetails(reqId, contract)


    def processBondContractDataMsg(self, fields):

        next(fields)
        version = 6;
        if self.serverVersion < MIN_SERVER_VER_SIZE_RULES:
            version = decode(int, fields)

        reqId = -1
        if version >= 3:
            reqId = decode(int, fields)

        contract = ContractDetails()
        contract.contract.symbol = decode(str, fields)
        contract.contract.secType = decode(str, fields)
        contract.cusip = decode(str, fields)
        contract.coupon = decode(int, fields)
        self.readLastTradeDate(fields, contract, True)
        contract.issueDate = decode(str, fields)
        contract.ratings = decode(str, fields)
        contract.bondType = decode(str, fields)
        contract.couponType = decode(str, fields)
        contract.convertible = decode(bool, fields)
        contract.callable = decode(bool, fields)
        contract.putable = decode(bool, fields)
        contract.descAppend = decode(str, fields)
        contract.contract.exchange = decode(str, fields)
        contract.contract.currency = decode(str, fields)
        contract.marketName = decode(str, fields)
        contract.contract.tradingClass = decode(str, fields)
        contract.contract.conId = decode(int, fields)
        contract.minTick = decode(float, fields)
        if self.serverVersion >= MIN_SERVER_VER_MD_SIZE_MULTIPLIER and self.serverVersion < MIN_SERVER_VER_SIZE_RULES:
            decode(int, fields) # mdSizeMultiplier - not used anymore
        contract.orderTypes = decode(str, fields)
        contract.validExchanges = decode(str, fields)
        contract.nextOptionDate = decode(str, fields) # ver 2 field
        contract.nextOptionType = decode(str, fields) # ver 2 field
        contract.nextOptionPartial = decode(bool, fields) # ver 2 field
        contract.notes = decode(str, fields) # ver 2 field
        if version >= 4:
            contract.longName = decode(str, fields)
        if version >= 6:
            contract.evRule = decode(str, fields)
            contract.evMultiplier = decode(int, fields)
        if version >= 5:
            secIdListCount = decode(int, fields)
            if secIdListCount > 0:
                contract.secIdList = []
                for _ in range(secIdListCount):
                    tagValue = TagValue()
                    tagValue.tag = decode(str, fields)
                    tagValue.value = decode(str, fields)
                    contract.secIdList.append(tagValue)

        if self.serverVersion >= MIN_SERVER_VER_AGG_GROUP:
            contract.aggGroup = decode(int, fields)

        if self.serverVersion >= MIN_SERVER_VER_MARKET_RULES:
            contract.marketRuleIds = decode(str, fields)

        if self.serverVersion >= MIN_SERVER_VER_SIZE_RULES:
            contract.minSize = decode(Decimal, fields)
            contract.sizeIncrement = decode(Decimal, fields)
            contract.suggestedSizeIncrement = decode(Decimal, fields)

        self.wrapper.bondContractDetails(reqId, contract)

    def processScannerDataMsg(self, fields):
        next(fields)
        decode(int, fields)
        reqId = decode(int, fields)

        numberOfElements = decode(int, fields)

        for _ in range(numberOfElements):
            data = ScanData()
            data.contract = ContractDetails()

            data.rank = decode(int, fields)
            data.contract.contract.conId = decode(int, fields) # ver 3 field
            data.contract.contract.symbol = decode(str, fields)
            data.contract.contract.secType = decode(str, fields)
            data.contract.contract.lastTradeDateOrContractMonth = decode(str, fields)
            data.contract.contract.strike = decode(float, fields)
            data.contract.contract.right = decode(str, fields)
            data.contract.contract.exchange = decode(str, fields)
            data.contract.contract.currency = decode(str, fields)
            data.contract.contract.localSymbol = decode(str, fields)
            data.contract.marketName = decode(str, fields)
            data.contract.contract.tradingClass = decode(str, fields)
            data.distance = decode(str, fields)
            data.benchmark = decode(str, fields)
            data.projection = decode(str, fields)
            data.legsStr = decode(str, fields)
            self.wrapper.scannerData(reqId, data.rank, data.contract,
                data.distance, data.benchmark, data.projection, data.legsStr)

        self.wrapper.scannerDataEnd(reqId)


    def processExecutionDataMsg(self, fields):
        next(fields)
        version = self.serverVersion

        if(self.serverVersion < MIN_SERVER_VER_LAST_LIQUIDITY):
            version = decode(int, fields)

        reqId = -1
        if version >= 7:
            reqId = decode(int, fields)

        orderId = decode(int, fields)

        # decode contract fields
        contract = Contract()
        contract.conId = decode(int, fields) # ver 5 field
        contract.symbol = decode(str, fields)
        contract.secType = decode(str, fields)
        contract.lastTradeDateOrContractMonth = decode(str, fields)
        contract.strike = decode(float, fields)
        contract.right = decode(str, fields)
        if version >= 9:
            contract.multiplier = decode(str, fields)
        contract.exchange = decode(str, fields)
        contract.currency = decode(str, fields)
        contract.localSymbol = decode(str, fields)
        if version >= 10:
            contract.tradingClass = decode(str, fields)

        # decode execution fields
        execution = Execution()
        execution.orderId = orderId
        execution.execId = decode(str, fields)
        execution.time = decode(str, fields)
        execution.acctNumber = decode(str, fields)
        execution.exchange = decode(str, fields)
        execution.side = decode(str, fields)
        execution.shares = decode(Decimal, fields)
        execution.price = decode(float, fields)
        execution.permId = decode(int, fields) # ver 2 field
        execution.clientId = decode(int, fields)  # ver 3 field
        execution.liquidation = decode(int, fields) # ver 4 field

        if version >= 6:
            execution.cumQty = decode(Decimal, fields)
            execution.avgPrice = decode(float, fields)

        if version >= 8:
            execution.orderRef = decode(str, fields)

        if version >= 9:
            execution.evRule = decode(str, fields)
            execution.evMultiplier = decode(float, fields)
        if self.serverVersion >= MIN_SERVER_VER_MODELS_SUPPORT:
            execution.modelCode = decode(str, fields)
        if self.serverVersion >= MIN_SERVER_VER_LAST_LIQUIDITY:
            execution.lastLiquidity = decode(int, fields)

        self.wrapper.execDetails(reqId, contract, execution)


    def processHistoricalDataMsg(self, fields):
        next(fields)

        if self.serverVersion < MIN_SERVER_VER_SYNT_REALTIME_BARS:
            decode(int, fields)

        reqId = decode(int, fields)
        startDateStr = decode(str, fields) # ver 2 field
        endDateStr = decode(str, fields) # ver 2 field

        itemCount = decode(int, fields)

        for _ in range(itemCount):
            bar = BarData()
            bar.date = decode(str, fields)
            bar.open = decode(float, fields)
            bar.high = decode(float, fields)
            bar.low = decode(float, fields)
            bar.close = decode(float, fields)
            bar.volume = decode(Decimal, fields)
            bar.wap = decode(Decimal, fields)

            if self.serverVersion < MIN_SERVER_VER_SYNT_REALTIME_BARS:
                decode(str, fields)

            bar.barCount = decode(int, fields) # ver 3 field

            self.wrapper.historicalData(reqId, bar)

        # send end of dataset marker
        self.wrapper.historicalDataEnd(reqId, startDateStr, endDateStr)

    def processHistoricalDataUpdateMsg(self, fields):
        next(fields)
        reqId = decode(int, fields)
        bar = BarData()
        bar.barCount = decode(int, fields)
        bar.date = decode(str, fields)
        bar.open = decode(float, fields)
        bar.close = decode(float, fields)
        bar.high = decode(float, fields)
        bar.low = decode(float, fields)
        bar.wap = decode(Decimal, fields)
        bar.volume = decode(Decimal, fields)
        self.wrapper.historicalDataUpdate(reqId, bar)

    def processRealTimeBarMsg(self, fields):
        next(fields)
        decode(int, fields)
        reqId = decode(int, fields)

        bar = RealTimeBar()
        bar.time = decode(int, fields)
        bar.open = decode(float, fields)
        bar.high = decode(float, fields)
        bar.low = decode(float, fields)
        bar.close = decode(float, fields)
        bar.volume = decode(Decimal, fields)
        bar.wap = decode(Decimal, fields)
        bar.count = decode(int, fields)

        self.wrapper.realtimeBar(reqId, bar.time, bar.open, bar.high, bar.low, bar.close, bar.volume, bar.wap, bar.count)

    def processTickOptionComputationMsg(self, fields):
        version = self.serverVersion
        tickAttrib = None
        optPrice = None
        pvDividend = None
        gamma = None
        vega = None
        theta = None
        undPrice = None

        next(fields)
        if self.serverVersion < MIN_SERVER_VER_PRICE_BASED_VOLATILITY:
            version = decode(int, fields)

        reqId = decode(int, fields)
        tickTypeInt = decode(int, fields)

        if self.serverVersion >= MIN_SERVER_VER_PRICE_BASED_VOLATILITY:
            tickAttrib = decode(int, fields)

        impliedVol = decode(float, fields)
        delta = decode(float, fields)

        if impliedVol < 0:    # -1 is the "not computed" indicator
            impliedVol = None
        if delta == -2: # -2 is the "not computed" indicator
            delta = None

        if version >= 6 or \
            tickTypeInt == TickTypeEnum.MODEL_OPTION or \
                        tickTypeInt == TickTypeEnum.DELAYED_MODEL_OPTION:

            optPrice = decode(float, fields)
            pvDividend = decode(float, fields)

            if optPrice == -1:    # -1 is the "not computed" indicator
                optPrice = None
            if pvDividend == -1:  # -1 is the "not computed" indicator
                pvDividend = None

        if version >= 6:
            gamma = decode(float, fields)
            vega = decode(float, fields)
            theta = decode(float, fields)
            undPrice = decode(float, fields)

            if gamma == -2:  # -2 is the "not yet computed" indicator
                gamma = None
            if vega == -2:    # -2 is the "not yet computed" indicator
                vega = None
            if theta == -2:  # -2 is the "not yet computed" indicator
                theta = None
            if undPrice == -1:             # -1 is the "not computed" indicator
                undPrice = None

        self.wrapper.tickOptionComputation(reqId, tickTypeInt, tickAttrib, impliedVol,
            delta, optPrice, pvDividend, gamma, vega, theta, undPrice)



    def processDeltaNeutralValidationMsg(self, fields):
        next(fields)
        decode(int, fields)
        reqId = decode(int, fields)

        deltaNeutralContract = DeltaNeutralContract()

        deltaNeutralContract.conId = decode(int, fields)
        deltaNeutralContract.delta = decode(float, fields)
        deltaNeutralContract.price = decode(float, fields)

        self.wrapper.deltaNeutralValidation(reqId, deltaNeutralContract)


    def processMarketDataTypeMsg(self, fields):
        next(fields)
        decode(int, fields)
        reqId = decode(int, fields)
        marketDataType = decode(int, fields)

        self.wrapper.marketDataType(reqId, marketDataType)

    def processCommissionReportMsg(self, fields):
        next(fields)
        decode(int, fields)

        commissionReport = CommissionReport()
        commissionReport.execId = decode(str, fields)
        commissionReport.commission = decode(float, fields)
        commissionReport.currency = decode(str, fields)
        commissionReport.realizedPNL = decode(float, fields)
        commissionReport.yield_ = decode(float, fields)
        commissionReport.yieldRedemptionDate = decode(int, fields)

        self.wrapper.commissionReport(commissionReport)


    def processPositionDataMsg(self, fields):
        next(fields)
        version = decode(int, fields)

        account = decode(str, fields)

        # decode contract fields
        contract = Contract()
        contract.conId = decode(int, fields)
        contract.symbol = decode(str, fields)
        contract.secType = decode(str, fields)
        contract.lastTradeDateOrContractMonth = decode(str, fields)
        contract.strike = decode(float, fields)
        contract.right = decode(str, fields)
        contract.multiplier = decode(str, fields)
        contract.exchange = decode(str, fields)
        contract.currency = decode(str, fields)
        contract.localSymbol = decode(str, fields)
        if version >= 2:
            contract.tradingClass = decode(str, fields)

        position = decode(Decimal, fields)

        avgCost = 0.
        if version >= 3:
            avgCost = decode(float, fields)

        self.wrapper.position(account, contract, position, avgCost)


    def processPositionMultiMsg(self, fields):
        next(fields)
        decode(int, fields)
        reqId = decode(int, fields)
        account = decode(str, fields)

        # decode contract fields
        contract = Contract()
        contract.conId = decode(int, fields)
        contract.symbol = decode(str, fields)
        contract.secType = decode(str, fields)
        contract.lastTradeDateOrContractMonth = decode(str, fields)
        contract.strike = decode(float, fields)
        contract.right = decode(str, fields)
        contract.multiplier = decode(str, fields)
        contract.exchange = decode(str, fields)
        contract.currency = decode(str, fields)
        contract.localSymbol = decode(str, fields)
        contract.tradingClass = decode(str, fields)
        position = decode(Decimal, fields)
        avgCost = decode(float, fields)
        modelCode = decode(str, fields)

        self.wrapper.positionMulti(reqId, account, modelCode, contract, position, avgCost)


    def processSecurityDefinitionOptionParameterMsg(self, fields):
        next(fields)

        reqId = decode(int, fields)
        exchange = decode(str, fields)
        underlyingConId = decode(int, fields)
        tradingClass = decode(str, fields)
        multiplier = decode(str, fields)

        expCount = decode(int, fields)
        expirations = set()
        for _ in range(expCount):
            expiration = decode(str, fields)
            expirations.add(expiration)

        strikeCount = decode(int, fields)
        strikes = set()
        for _ in range(strikeCount):
            strike = decode(float, fields)
            strikes.add(strike)

        self.wrapper.securityDefinitionOptionParameter(reqId, exchange,
            underlyingConId, tradingClass, multiplier, expirations, strikes)


    def processSecurityDefinitionOptionParameterEndMsg(self, fields):
        next(fields)

        reqId = decode(int, fields)
        self.wrapper.securityDefinitionOptionParameterEnd(reqId)


    def processSoftDollarTiersMsg(self, fields):
        next(fields)

        reqId = decode(int, fields)
        nTiers = decode(int, fields)

        tiers = []
        for _ in range(nTiers):
                tier = SoftDollarTier()
                tier.name = decode(str, fields)
                tier.val = decode(str, fields)
                tier.displayName = decode(str, fields)
                tiers.append(tier)

        self.wrapper.softDollarTiers(reqId, tiers)


    def processFamilyCodesMsg(self, fields):
        next(fields)

        nFamilyCodes = decode(int, fields)
        familyCodes = []
        for _ in range(nFamilyCodes):
            famCode = FamilyCode()
            famCode.accountID = decode(str, fields)
            famCode.familyCodeStr = decode(str, fields)
            familyCodes.append(famCode)

        self.wrapper.familyCodes(familyCodes)


    def processSymbolSamplesMsg(self, fields):
        next(fields)

        reqId = decode(int, fields)
        nContractDescriptions = decode(int, fields)
        contractDescriptions = []
        for _ in range(nContractDescriptions):
            conDesc = ContractDescription()
            conDesc.contract.conId = decode(int, fields)
            conDesc.contract.symbol = decode(str, fields)
            conDesc.contract.secType = decode(str, fields)
            conDesc.contract.primaryExchange = decode(str, fields)
            conDesc.contract.currency = decode(str, fields)

            nDerivativeSecTypes = decode(int, fields)
            conDesc.derivativeSecTypes = []
            for _ in range(nDerivativeSecTypes):
                derivSecType = decode(str, fields)
                conDesc.derivativeSecTypes.append(derivSecType)
            contractDescriptions.append(conDesc)

        self.wrapper.symbolSamples(reqId, contractDescriptions)

    def processSmartComponents(self,fields):
        next(fields)
        reqId = decode(int, fields)
        n = decode(int, fields)

        smartComponentMap = []
        for _ in range(n):
            smartComponent = SmartComponent()
            smartComponent.bitNumber = decode(int, fields)
            smartComponent.exchange = decode(str, fields)
            smartComponent.exchangeLetter = decode(str, fields)
            smartComponentMap.append(smartComponent)

        self.wrapper.smartComponents(reqId, smartComponentMap)

    def processTickReqParams(self,fields):
        next(fields)
        tickerId = decode(int, fields)
        minTick = decode(float, fields)
        bboExchange = decode(str, fields)
        snapshotPermissions = decode(int, fields)
        self.wrapper.tickReqParams(tickerId, minTick, bboExchange, snapshotPermissions)

    def processMktDepthExchanges(self,fields):
        next(fields)
        depthMktDataDescriptions = []
        nDepthMktDataDescriptions = decode(int, fields)

        if nDepthMktDataDescriptions > 0:
            for _ in range(nDepthMktDataDescriptions):
                desc = DepthMktDataDescription()
                desc.exchange = decode(str, fields)
                desc.secType = decode(str, fields)
                if self.serverVersion >= MIN_SERVER_VER_SERVICE_DATA_TYPE:
                    desc.listingExch = decode(str, fields)
                    desc.serviceDataType = decode(str, fields)
                    desc.aggGroup = decode(int, fields)
                else:
                    decode(int,fields) #boolean notSuppIsL2
                depthMktDataDescriptions.append(desc)

        self.wrapper.mktDepthExchanges(depthMktDataDescriptions)

    def processHeadTimestamp(self,fields):
        next(fields)
        reqId = decode(int, fields)
        headTimestamp = decode(str, fields)
        self.wrapper.headTimestamp(reqId,headTimestamp)

    def processTickNews(self,fields):
        next(fields)
        tickerId = decode( int, fields)
        timeStamp = decode(int, fields)
        providerCode = decode(str, fields)
        articleId = decode(str, fields)
        headline = decode(str, fields)
        extraData = decode(str, fields)
        self.wrapper.tickNews(tickerId, timeStamp, providerCode, articleId, headline, extraData)

    def processNewsProviders(self,fields):
        next(fields)
        newsProviders = []
        nNewsProviders = decode(int, fields)
        if nNewsProviders > 0:
            for _ in range(nNewsProviders):
                provider = NewsProvider()
                provider.code = decode(str, fields)
                provider.name = decode(str, fields)
                newsProviders.append(provider)

        self.wrapper.newsProviders(newsProviders)

    def processNewsArticle(self,fields):
        next(fields)
        reqId = decode(int, fields)
        articleType = decode(int, fields)
        articleText = decode(str, fields)
        self.wrapper.newsArticle(reqId, articleType, articleText)

    def processHistoricalNews(self,fields):
        next(fields)
        requestId = decode(int, fields)
        time = decode(str, fields)
        providerCode = decode(str, fields)
        articleId = decode(str, fields)
        headline = decode(str, fields)
        self.wrapper.historicalNews(requestId, time, providerCode, articleId, headline)

    def processHistoricalNewsEnd(self,fields):
        next(fields)
        reqId = decode(int, fields)
        hasMore = decode(bool, fields)
        self.wrapper.historicalNewsEnd(reqId, hasMore)

    def processHistogramData(self,fields):
        next(fields)
        reqId = decode(int, fields)
        numPoints = decode(int, fields)

        histogram = []
        for _ in range(numPoints):
            dataPoint = HistogramData()
            dataPoint.price = decode(float,fields)
            dataPoint.size = decode(Decimal,fields)
            histogram.append(dataPoint)

        self.wrapper.histogramData(reqId, histogram)

    def processRerouteMktDataReq(self, fields):
        next(fields)
        reqId = decode(int, fields)
        conId = decode(int, fields)
        exchange = decode(str, fields)

        self.wrapper.rerouteMktDataReq(reqId, conId, exchange)

    def processRerouteMktDepthReq(self, fields):
        next(fields)
        reqId = decode(int, fields)
        conId = decode(int, fields)
        exchange = decode(str, fields)

        self.wrapper.rerouteMktDepthReq(reqId, conId, exchange)

    def processMarketRuleMsg(self, fields):
        next(fields)
        marketRuleId = decode(int, fields)

        nPriceIncrements = decode(int, fields)
        priceIncrements = []

        if nPriceIncrements > 0:
            for _ in range(nPriceIncrements):
                prcInc = PriceIncrement()
                prcInc.lowEdge = decode(float, fields)
                prcInc.increment = decode(float, fields)
                priceIncrements.append(prcInc)

        self.wrapper.marketRule(marketRuleId, priceIncrements)

    def processPnLMsg(self, fields):
        next(fields)
        reqId = decode(int, fields)
        dailyPnL = decode(float, fields)
        unrealizedPnL = None
        realizedPnL = None

        if self.serverVersion >= MIN_SERVER_VER_UNREALIZED_PNL:
            unrealizedPnL = decode(float, fields)

        if self.serverVersion >= MIN_SERVER_VER_REALIZED_PNL:
            realizedPnL = decode(float, fields)

        self.wrapper.pnl(reqId, dailyPnL, unrealizedPnL, realizedPnL)

    def processPnLSingleMsg(self, fields):
        next(fields)
        reqId = decode(int, fields)
        pos = decode(Decimal, fields)
        dailyPnL = decode(float, fields)
        unrealizedPnL = None
        realizedPnL = None

        if self.serverVersion >= MIN_SERVER_VER_UNREALIZED_PNL:
            unrealizedPnL = decode(float, fields)

        if self.serverVersion >= MIN_SERVER_VER_REALIZED_PNL:
            realizedPnL = decode(float, fields)

        value = decode(float, fields)

        self.wrapper.pnlSingle(reqId, pos, dailyPnL, unrealizedPnL, realizedPnL, value)

    def processHistoricalTicks(self, fields):
        next(fields)
        reqId = decode(int, fields)
        tickCount = decode(int, fields)

        ticks = []

        for _ in range(tickCount):
            historicalTick = HistoricalTick()
            historicalTick.time = decode(int, fields)
            next(fields) # for consistency
            historicalTick.price = decode(float, fields)
            historicalTick.size = decode(Decimal, fields)
            ticks.append(historicalTick)

        done = decode(bool, fields)

        self.wrapper.historicalTicks(reqId, ticks, done)

    def processHistoricalTicksBidAsk(self, fields):
        next(fields)
        reqId = decode(int, fields)
        tickCount = decode(int, fields)

        ticks = []

        for _ in range(tickCount):
            historicalTickBidAsk = HistoricalTickBidAsk()
            historicalTickBidAsk.time = decode(int, fields)
            mask = decode(int, fields)
            tickAttribBidAsk = TickAttribBidAsk()
            tickAttribBidAsk.askPastHigh = mask & 1 != 0
            tickAttribBidAsk.bidPastLow = mask & 2 != 0
            historicalTickBidAsk.tickAttribBidAsk = tickAttribBidAsk
            historicalTickBidAsk.priceBid = decode(float, fields)
            historicalTickBidAsk.priceAsk = decode(float, fields)
            historicalTickBidAsk.sizeBid = decode(Decimal, fields)
            historicalTickBidAsk.sizeAsk = decode(Decimal, fields)
            ticks.append(historicalTickBidAsk)

        done = decode(bool, fields)

        self.wrapper.historicalTicksBidAsk(reqId, ticks, done)

    def processHistoricalTicksLast(self, fields):
        next(fields)
        reqId = decode(int, fields)
        tickCount = decode(int, fields)

        ticks = []

        for _ in range(tickCount):
            historicalTickLast = HistoricalTickLast()
            historicalTickLast.time = decode(int, fields)
            mask = decode(int, fields)
            tickAttribLast = TickAttribLast()
            tickAttribLast.pastLimit = mask & 1 != 0
            tickAttribLast.unreported = mask & 2 != 0
            historicalTickLast.tickAttribLast = tickAttribLast
            historicalTickLast.price = decode(float, fields)
            historicalTickLast.size = decode(Decimal, fields)
            historicalTickLast.exchange = decode(str, fields)
            historicalTickLast.specialConditions = decode(str, fields)
            ticks.append(historicalTickLast)

        done = decode(bool, fields)

        self.wrapper.historicalTicksLast(reqId, ticks, done)

    def processTickByTickMsg(self, fields):
        next(fields)
        reqId = decode(int, fields)
        tickType = decode(int, fields)
        time = decode(int, fields)

        if tickType == 0:
            # None
            pass
        elif tickType == 1 or tickType == 2:
            # Last or AllLast
            price = decode(float, fields)
            size = decode(Decimal, fields)
            mask = decode(int, fields)

            tickAttribLast = TickAttribLast()
            tickAttribLast.pastLimit = mask & 1 != 0
            tickAttribLast.unreported = mask & 2 != 0
            exchange = decode(str, fields)
            specialConditions = decode(str, fields)

            self.wrapper.tickByTickAllLast(reqId, tickType, time, price, size, tickAttribLast,
                                           exchange, specialConditions)
        elif tickType == 3:
            # BidAsk
            bidPrice = decode(float, fields)
            askPrice = decode(float, fields)
            bidSize = decode(Decimal, fields)
            askSize = decode(Decimal, fields)
            mask = decode(int, fields)
            tickAttribBidAsk = TickAttribBidAsk()
            tickAttribBidAsk.bidPastLow = mask & 1 != 0
            tickAttribBidAsk.askPastHigh = mask & 2 != 0

            self.wrapper.tickByTickBidAsk(reqId, time, bidPrice, askPrice, bidSize,
                                          askSize, tickAttribBidAsk)
        elif tickType == 4:
            # MidPoint
            midPoint = decode(float, fields)

            self.wrapper.tickByTickMidPoint(reqId, time, midPoint)

    def processOrderBoundMsg(self, fields):
        next(fields)
        reqId = decode(int, fields)
        apiClientId = decode(int, fields)
        apiOrderId = decode(int, fields)

        self.wrapper.orderBound(reqId, apiClientId, apiOrderId)

    def processMarketDepthMsg(self, fields):
        next(fields)
        decode(int, fields)
        reqId = decode(int, fields)

        position = decode(int, fields)
        operation = decode(int, fields)
        side = decode(int, fields)
        price = decode(float, fields)
        size = decode(Decimal, fields)

        self.wrapper.updateMktDepth(reqId, position, operation, side, price, size)

    def processMarketDepthL2Msg(self, fields):
        next(fields)
        decode(int, fields)
        reqId = decode(int, fields)

        position = decode(int, fields)
        marketMaker = decode(str, fields)
        operation = decode(int, fields)
        side = decode(int, fields)
        price = decode(float, fields)
        size = decode(Decimal, fields)
        isSmartDepth = False

        if self.serverVersion >= MIN_SERVER_VER_SMART_DEPTH:
            isSmartDepth = decode(bool, fields)

        self.wrapper.updateMktDepthL2(reqId, position, marketMaker,
                        operation, side, price, size, isSmartDepth)


    def processCompletedOrderMsg(self, fields):
        next(fields)
        
        order = Order()
        contract = Contract()
        orderState = OrderState()

        OrderDecoder.__init__(self, contract, order, orderState, UNSET_INTEGER, self.serverVersion)

        # read contract fields
        OrderDecoder.decodeContractFields(self, fields)

        # read order fields
        OrderDecoder.decodeAction(self, fields)
        OrderDecoder.decodeTotalQuantity(self, fields)
        OrderDecoder.decodeOrderType(self, fields)
        OrderDecoder.decodeLmtPrice(self, fields)
        OrderDecoder.decodeAuxPrice(self, fields)
        OrderDecoder.decodeTIF(self, fields)
        OrderDecoder.decodeOcaGroup(self, fields)
        OrderDecoder.decodeAccount(self, fields)
        OrderDecoder.decodeOpenClose(self, fields)
        OrderDecoder.decodeOrigin(self, fields)
        OrderDecoder.decodeOrderRef(self, fields)
        OrderDecoder.decodePermId(self, fields)
        OrderDecoder.decodeOutsideRth(self, fields)
        OrderDecoder.decodeHidden(self, fields)
        OrderDecoder.decodeDiscretionaryAmt(self, fields)
        OrderDecoder.decodeGoodAfterTime(self, fields)
        OrderDecoder.decodeFAParams(self, fields)
        OrderDecoder.decodeModelCode(self, fields)
        OrderDecoder.decodeGoodTillDate(self, fields)
        OrderDecoder.decodeRule80A(self, fields)
        OrderDecoder.decodePercentOffset(self, fields)
        OrderDecoder.decodeSettlingFirm(self, fields)
        OrderDecoder.decodeShortSaleParams(self, fields)
        OrderDecoder.decodeBoxOrderParams(self, fields)
        OrderDecoder.decodePegToStkOrVolOrderParams(self, fields)
        OrderDecoder.decodeDisplaySize(self, fields)
        OrderDecoder.decodeSweepToFill(self, fields)
        OrderDecoder.decodeAllOrNone(self, fields)
        OrderDecoder.decodeMinQty(self, fields)
        OrderDecoder.decodeOcaType(self, fields)
        OrderDecoder.decodeTriggerMethod(self, fields)
        OrderDecoder.decodeVolOrderParams(self, fields, False)
        OrderDecoder.decodeTrailParams(self, fields)
        OrderDecoder.decodeComboLegs(self, fields)
        OrderDecoder.decodeSmartComboRoutingParams(self, fields)
        OrderDecoder.decodeScaleOrderParams(self, fields)
        OrderDecoder.decodeHedgeParams(self, fields)
        OrderDecoder.decodeClearingParams(self, fields)
        OrderDecoder.decodeNotHeld(self, fields)
        OrderDecoder.decodeDeltaNeutral(self, fields)
        OrderDecoder.decodeAlgoParams(self, fields)
        OrderDecoder.decodeSolicited(self, fields)
        OrderDecoder.decodeOrderStatus(self, fields)
        OrderDecoder.decodeVolRandomizeFlags(self, fields)
        OrderDecoder.decodePegToBenchParams(self, fields)
        OrderDecoder.decodeConditions(self, fields)
        OrderDecoder.decodeStopPriceAndLmtPriceOffset(self, fields)
        OrderDecoder.decodeCashQty(self, fields)
        OrderDecoder.decodeDontUseAutoPriceForHedge(self, fields)
        OrderDecoder.decodeIsOmsContainers(self, fields)
        OrderDecoder.decodeAutoCancelDate(self, fields)
        OrderDecoder.decodeFilledQuantity(self, fields)
        OrderDecoder.decodeRefFuturesConId(self, fields)
        OrderDecoder.decodeAutoCancelParent(self, fields)
        OrderDecoder.decodeShareholder(self, fields)
        OrderDecoder.decodeImbalanceOnly(self, fields)
        OrderDecoder.decodeRouteMarketableToBbo(self, fields)
        OrderDecoder.decodeParentPermId(self, fields)
        OrderDecoder.decodeCompletedTime(self, fields)
        OrderDecoder.decodeCompletedStatus(self, fields)

        self.wrapper.completedOrder(contract, order, orderState)

    def processCompletedOrdersEndMsg(self, fields):
        next(fields)

        self.wrapper.completedOrdersEnd()
        
    def processReplaceFAEndMsg(self, fields):
        next(fields)
        reqId = decode(int, fields)
        text = decode(str, fields)

        self.wrapper.replaceFAEnd(reqId, text)

    def processWshMetaDataMsg(self, fields):
        next(fields)
        reqId = decode(int, fields)
        dataJson = decode(str, fields)

        self.wrapper.wshMetaData(reqId, dataJson)

    def processWshEventDataMsg(self, fields):
        next(fields)
        reqId = decode(int, fields)
        dataJson = decode(str, fields)

        self.wrapper.wshEventData(reqId, dataJson)

    ######################################################################

    def readLastTradeDate(self, fields, contract: ContractDetails, isBond: bool):
        lastTradeDateOrContractMonth = decode(str, fields)
        if lastTradeDateOrContractMonth is not None:
            splitted = lastTradeDateOrContractMonth.split()
            if len(splitted) > 0:
                if isBond:
                    contract.maturity = splitted[0]
                else:
                    contract.contract.lastTradeDateOrContractMonth = splitted[0]

            if len(splitted) > 1:
                contract.lastTradeTime = splitted[1]

            if isBond and len(splitted) > 2:
                contract.timeZoneId = splitted[2]

    ######################################################################

    def discoverParams(self):
        meth2handleInfo = {}
        for handleInfo in self.msgId2handleInfo.values():
            meth2handleInfo[handleInfo.wrapperMeth] = handleInfo

        methods = inspect.getmembers(EWrapper, inspect.isfunction)
        for (_, meth) in methods:
            #logger.debug("meth %s", name)
            sig = inspect.signature(meth)
            handleInfo = meth2handleInfo.get(meth, None)
            if handleInfo is not None:
                handleInfo.wrapperParams = sig.parameters

            #for (pname, param) in sig.parameters.items():
            #     logger.debug("\tparam %s %s %s", pname, param.name, param.annotation)


    def printParams(self):
        for (_, handleInfo) in self.msgId2handleInfo.items():
            if handleInfo.wrapperMeth is not None:
                logger.debug("meth %s", handleInfo.wrapperMeth.__name__)
                if handleInfo.wrapperParams is not None:
                    for (pname, param) in handleInfo.wrapperParams.items():
                        logger.debug("\tparam %s %s %s", pname, param.name, param.annotation)


    def interpretWithSignature(self, fields, handleInfo):
        if handleInfo.wrapperParams is None:
            logger.debug("%s: no param info in %s", fields, handleInfo)
            return

        nIgnoreFields = 2 #bypass msgId and versionId faster this way
        if len(fields) - nIgnoreFields != len(handleInfo.wrapperParams) - 1:
            logger.error("diff len fields and params %d %d for fields: %s and handleInfo: %s",
                         len(fields), len(handleInfo.wrapperParams), fields,
                         handleInfo)
            return

        fieldIdx = nIgnoreFields
        args = []
        for (pname, param) in handleInfo.wrapperParams.items():
            if pname != "self":
                logger.debug("field %s ", fields[fieldIdx])
                try:
                    arg = fields[fieldIdx].decode('unicode-escape' if self.serverVersion >= MIN_SERVER_VER_ENCODE_MSG_ASCII7 else 'UTF-8')
                except UnicodeDecodeError:
                    arg = fields[fieldIdx].decode('latin-1')
                logger.debug("arg %s type %s", arg, param.annotation)
                if param.annotation is int:
                    arg = int(arg)
                elif param.annotation is float:
                    arg = float(arg)
                elif param.annotation is Decimal:
                    if arg is None or len(arg) == 0:
                        return UNSET_DECIMAL
                    else:
                        return Decimal(arg)

                args.append(arg)
                fieldIdx += 1

        method = getattr(self.wrapper, handleInfo.wrapperMeth.__name__)
        logger.debug("calling %s with %s %s", method, self.wrapper, args)
        method(*args)

    def interpret(self, fields):
        if len(fields) == 0:
            logger.debug("no fields")
            return

        sMsgId = fields[0]
        nMsgId = int(sMsgId)

        handleInfo = self.msgId2handleInfo.get(nMsgId, None)

        if handleInfo is None:
            logger.debug("%s: no handleInfo", fields)
            return

        try:
            if handleInfo.wrapperMeth is not None:
                logger.debug("In interpret(), handleInfo: %s", handleInfo)
                self.interpretWithSignature(fields, handleInfo)
            elif handleInfo.processMeth is not None:
                handleInfo.processMeth(self, iter(fields))
        except BadMessage:
                theBadMsg = ",".join(fields)
                self.wrapper.error(NO_VALID_ID, BAD_MESSAGE.code(),
                                   BAD_MESSAGE.msg() + theBadMsg)
                raise


    msgId2handleInfo = {
        IN.TICK_PRICE: HandleInfo(proc=processTickPriceMsg),
        IN.TICK_SIZE: HandleInfo(proc=processTickSizeMsg),
        IN.ORDER_STATUS: HandleInfo(proc=processOrderStatusMsg),
        IN.ERR_MSG: HandleInfo(wrap=EWrapper.error),
        IN.OPEN_ORDER: HandleInfo(proc=processOpenOrder),
        IN.ACCT_VALUE: HandleInfo(wrap=EWrapper.updateAccountValue),
        IN.PORTFOLIO_VALUE: HandleInfo(proc=processPortfolioValueMsg),
        IN.ACCT_UPDATE_TIME: HandleInfo(wrap=EWrapper.updateAccountTime),
        IN.NEXT_VALID_ID: HandleInfo(wrap=EWrapper.nextValidId, ),
        IN.CONTRACT_DATA: HandleInfo(proc=processContractDataMsg),
        IN.EXECUTION_DATA: HandleInfo(proc=processExecutionDataMsg),
        IN.MARKET_DEPTH: HandleInfo(proc=processMarketDepthMsg),
        IN.MARKET_DEPTH_L2: HandleInfo(proc=processMarketDepthL2Msg),
        IN.NEWS_BULLETINS: HandleInfo(wrap=EWrapper.updateNewsBulletin),
        IN.MANAGED_ACCTS: HandleInfo(wrap=EWrapper.managedAccounts),
        IN.RECEIVE_FA: HandleInfo(wrap=EWrapper.receiveFA),
        IN.HISTORICAL_DATA: HandleInfo(proc=processHistoricalDataMsg),
        IN.HISTORICAL_DATA_UPDATE: HandleInfo(proc=processHistoricalDataUpdateMsg),
        IN.BOND_CONTRACT_DATA: HandleInfo(proc=processBondContractDataMsg),
        IN.SCANNER_PARAMETERS: HandleInfo(wrap=EWrapper.scannerParameters),
        IN.SCANNER_DATA: HandleInfo(proc=processScannerDataMsg),
        IN.TICK_OPTION_COMPUTATION: HandleInfo(proc=processTickOptionComputationMsg),
        IN.TICK_GENERIC: HandleInfo(wrap=EWrapper.tickGeneric),
        IN.TICK_STRING: HandleInfo(wrap=EWrapper.tickString),
        IN.TICK_EFP: HandleInfo(wrap=EWrapper.tickEFP),
        IN.CURRENT_TIME: HandleInfo(wrap=EWrapper.currentTime),
        IN.REAL_TIME_BARS: HandleInfo(proc=processRealTimeBarMsg),
        IN.FUNDAMENTAL_DATA: HandleInfo(wrap=EWrapper.fundamentalData),
        IN.CONTRACT_DATA_END: HandleInfo(wrap=EWrapper.contractDetailsEnd),
        IN.OPEN_ORDER_END: HandleInfo(wrap=EWrapper.openOrderEnd),
        IN.ACCT_DOWNLOAD_END: HandleInfo(wrap=EWrapper.accountDownloadEnd),
        IN.EXECUTION_DATA_END: HandleInfo(wrap=EWrapper.execDetailsEnd),
        IN.DELTA_NEUTRAL_VALIDATION: HandleInfo(proc=processDeltaNeutralValidationMsg),
        IN.TICK_SNAPSHOT_END: HandleInfo(wrap=EWrapper.tickSnapshotEnd),
        IN.MARKET_DATA_TYPE: HandleInfo(wrap=EWrapper.marketDataType),
        IN.COMMISSION_REPORT: HandleInfo(proc=processCommissionReportMsg),
        IN.POSITION_DATA: HandleInfo(proc=processPositionDataMsg),
        IN.POSITION_END: HandleInfo(wrap=EWrapper.positionEnd),
        IN.ACCOUNT_SUMMARY: HandleInfo(wrap=EWrapper.accountSummary),
        IN.ACCOUNT_SUMMARY_END: HandleInfo(wrap=EWrapper.accountSummaryEnd),
        IN.VERIFY_MESSAGE_API: HandleInfo(wrap=EWrapper.verifyMessageAPI),
        IN.VERIFY_COMPLETED: HandleInfo(wrap=EWrapper.verifyCompleted),
        IN.DISPLAY_GROUP_LIST: HandleInfo(wrap=EWrapper.displayGroupList),
        IN.DISPLAY_GROUP_UPDATED: HandleInfo(wrap=EWrapper.displayGroupUpdated),
        IN.VERIFY_AND_AUTH_MESSAGE_API: HandleInfo(wrap=EWrapper.verifyAndAuthMessageAPI),
        IN.VERIFY_AND_AUTH_COMPLETED: HandleInfo(wrap=EWrapper.verifyAndAuthCompleted),
        IN.POSITION_MULTI: HandleInfo(proc=processPositionMultiMsg),
        IN.POSITION_MULTI_END: HandleInfo(wrap=EWrapper.positionMultiEnd),
        IN.ACCOUNT_UPDATE_MULTI: HandleInfo(wrap=EWrapper.accountUpdateMulti),
        IN.ACCOUNT_UPDATE_MULTI_END: HandleInfo(wrap=EWrapper.accountUpdateMultiEnd),
        IN.SECURITY_DEFINITION_OPTION_PARAMETER: HandleInfo(proc=processSecurityDefinitionOptionParameterMsg),
        IN.SECURITY_DEFINITION_OPTION_PARAMETER_END: HandleInfo(proc=processSecurityDefinitionOptionParameterEndMsg),
        IN.SOFT_DOLLAR_TIERS: HandleInfo(proc=processSoftDollarTiersMsg),
        IN.FAMILY_CODES: HandleInfo(proc=processFamilyCodesMsg),
        IN.SYMBOL_SAMPLES: HandleInfo(proc=processSymbolSamplesMsg),
        IN.SMART_COMPONENTS: HandleInfo(proc=processSmartComponents),
        IN.TICK_REQ_PARAMS: HandleInfo(proc=processTickReqParams),
        IN.MKT_DEPTH_EXCHANGES: HandleInfo(proc=processMktDepthExchanges),
        IN.HEAD_TIMESTAMP: HandleInfo(proc=processHeadTimestamp),
        IN.TICK_NEWS: HandleInfo(proc=processTickNews),
        IN.NEWS_PROVIDERS: HandleInfo(proc=processNewsProviders),
        IN.NEWS_ARTICLE: HandleInfo(proc=processNewsArticle),
        IN.HISTORICAL_NEWS: HandleInfo(proc=processHistoricalNews),
        IN.HISTORICAL_NEWS_END: HandleInfo(proc=processHistoricalNewsEnd),
        IN.HISTOGRAM_DATA: HandleInfo(proc=processHistogramData),
        IN.REROUTE_MKT_DATA_REQ: HandleInfo(proc=processRerouteMktDataReq),
        IN.REROUTE_MKT_DEPTH_REQ: HandleInfo(proc=processRerouteMktDepthReq),
        IN.MARKET_RULE: HandleInfo(proc=processMarketRuleMsg),
        IN.PNL: HandleInfo(proc=processPnLMsg),
        IN.PNL_SINGLE: HandleInfo(proc=processPnLSingleMsg),
        IN.HISTORICAL_TICKS: HandleInfo(proc=processHistoricalTicks),
        IN.HISTORICAL_TICKS_BID_ASK: HandleInfo(proc=processHistoricalTicksBidAsk),
        IN.HISTORICAL_TICKS_LAST: HandleInfo(proc=processHistoricalTicksLast),
        IN.TICK_BY_TICK: HandleInfo(proc=processTickByTickMsg),
        IN.ORDER_BOUND: HandleInfo(proc=processOrderBoundMsg),
        IN.COMPLETED_ORDER: HandleInfo(proc=processCompletedOrderMsg),
        IN.COMPLETED_ORDERS_END: HandleInfo(proc=processCompletedOrdersEndMsg),
        IN.REPLACE_FA_END: HandleInfo(proc=processReplaceFAEndMsg),
        IN.WSH_META_DATA: HandleInfo(proc=processWshMetaDataMsg),
        IN.WSH_EVENT_DATA: HandleInfo(proc=processWshEventDataMsg)
}






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 13 18:43:54 2021

@author: ron
"""

from abc import ABC, abstractmethod
from datetime import datetime, timezone


class AbstractStrategy(ABC):
    """
    AbstractStrategy is an abstract base class 
    for all strategy class
    ------------------
    """
    @abstractmethod
    def calculate_signal(self, bar_event):
        pass
    
    
    def save_to_db_signal(self, signal_event):
        query = ("INSERT INTO SignalEvent (Datetime, Ticker, Order_Type, Direction, Closing_Price, Entry_Price)"
                "VALUES (%s, %s, %s, %s, %s, %s)"
        )
        timestamp_converted = datetime.fromtimestamp(signal_event.timestamp, timezone.utc)
        data = (timestamp_converted, signal_event.ticker, signal_event.order_type.name, signal_event.direction.name, 
                signal_event.bar_closing_price, signal_event.entry_price)
        cursor = self.db_client.cursor()
        cursor.execute(query, data)
        self.db_client.commit()  
    
    





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 16 20:16:45 2021

@author: ron
"""

from .base import AbstractTechnicalIndicator


class RSI(AbstractTechnicalIndicator):
    
    def __init__(self, period):
        self.period = period
        self.prev_close = None
        self.lastest_val = 0
        self.data_store = []
        self.avg_gain = 0
        self.avg_loss = 0
        self.ready = False
        self.first_rsi_count = 0        


    def update(self, price):
        if self.prev_close is not None:
            gain, loss = self._gain_loss(price)
            #need data for calculating first rsi...
            if self.first_rsi_count < self.period:
                self.avg_gain += gain            
                self.avg_loss += loss
                    
                self.first_rsi_count += 1
                if not self.ready and self.first_rsi_count == self.period:  
                    self.ready = True
                    
            
            #data is engough for calculating rsi...
            if self.ready:
                #averge gain and loss for the first rsi
                if self.first_rsi_count == self.period:
                    self.avg_gain = self.avg_gain / self.period
                    self.avg_loss = self.avg_loss / self.period
                    self.first_rsi_count += 1
                #average gain and loss for subsequent rsi    
                else:
                    self.avg_gain = (self.avg_gain * (self.period - 1) + gain) / self.period
                    self.avg_loss = (self.avg_loss * (self.period - 1) + loss) / self.period   
                    
                try:            
                    rsi = 100 - (100 / (1 + (self.avg_gain / self.avg_loss)))                    
                #no loss
                except ZeroDivisionError:
                    rsi = 100
                
                self.lastest_val = rsi
                self.data_store.append(rsi)
        self.prev_close = price
                                                     

    def _gain_loss(self, price):
        cur_gain_loss = price - self.prev_close
        if cur_gain_loss > 0:
            return cur_gain_loss, 0
        else:
            return 0, abs(cur_gain_loss)
        
                
    def __bool__(self):
        return self.ready
                    
            
                
                
                    
            
            
            
        


"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

"""
Simple class mapping a tag to a value. Both of them are strings. 
They are used in a list to convey extra info with the requests.
"""

from .object_implem import Object


class TagValue(Object):
    def __init__(self, tag:str=None, value:str=None):
        self.tag = str(tag)
        self.value = str(value)

    def __str__(self):
        # this is not only used for Python dump but when encoding to send
        # so don't change it lightly !
        return "%s=%s;" % (self.tag, self.value)



class Calendlar:
    def __init__(self):
        self.holidays = set()

    def add_holiday(self, date):
        #date: a date object
        self.holidays.add(date)

    def remove_holiday(self, date):
        #date: a date object
        self.holidays.remove(date)

    def is_holiday(self, date):
        #date: a date object
        return date in self.holidays

    def is_weekday(self, date):
        #date: a date object
        return date not in self.holidays



"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


from .common import UNSET_INTEGER, UNSET_DOUBLE, UNSET_DECIMAL
from .object_implem import Object
from .softdollartier import SoftDollarTier

# enum Origin
(CUSTOMER, FIRM, UNKNOWN) = range(3)

# enum AuctionStrategy
(AUCTION_UNSET, AUCTION_MATCH,
 AUCTION_IMPROVEMENT, AUCTION_TRANSPARENT) = range(4)


class OrderComboLeg(Object):
    def __init__(self):
        self.price = UNSET_DOUBLE  # type: float

    def __str__(self):
        return "%f" % self.price


class Order(Object):
    def __init__(self):
        self.softDollarTier = SoftDollarTier("", "", "")
        # order identifier
        self.orderId  = 0
        self.clientId = 0
        self.permId   = 0

        # main order fields
        self.action = ""
        self.totalQuantity = UNSET_DECIMAL
        self.orderType = ""
        self.lmtPrice      = UNSET_DOUBLE
        self.auxPrice      = UNSET_DOUBLE

        # extended order fields
        self.tif = ""                # "Time in Force" - DAY, GTC, etc.
        self.activeStartTime = ""   # for GTC orders
        self.activeStopTime = ""    # for GTC orders
        self.ocaGroup = ""          # one cancels all group name
        self.ocaType        = 0     # 1 = CANCEL_WITH_BLOCK, 2 = REDUCE_WITH_BLOCK, 3 = REDUCE_NON_BLOCK
        self.orderRef       = ""
        self.transmit       = True  # if false, order will be created but not transmited
        self.parentId       = 0     # Parent order Id, to associate Auto STP or TRAIL orders with the original order.
        self.blockOrder     = False
        self.sweepToFill    = False
        self.displaySize    = 0
        self.triggerMethod  = 0     # 0=Default, 1=Double_Bid_Ask, 2=Last, 3=Double_Last, 4=Bid_Ask, 7=Last_or_Bid_Ask, 8=Mid-point
        self.outsideRth     = False
        self.hidden         = False
        self.goodAfterTime       = ""   # Format: 20060505 08:00:00 {time zone}
        self.goodTillDate        = ""   # Format: 20060505 08:00:00 {time zone}
        self.rule80A             = ""   # Individual = 'I', Agency = 'A', AgentOtherMember = 'W', IndividualPTIA = 'J', AgencyPTIA = 'U', AgentOtherMemberPTIA = 'M', IndividualPT = 'K', AgencyPT = 'Y', AgentOtherMemberPT = 'N'
        self.allOrNone      = False
        self.minQty         = UNSET_INTEGER  # type: int
        self.percentOffset  = UNSET_DOUBLE  # type: float  # REL orders only
        self.overridePercentageConstraints = False
        self.trailStopPrice = UNSET_DOUBLE  # type: float
        self.trailingPercent = UNSET_DOUBLE # type: float  # TRAILLIMIT orders only

        # financial advisors only
        self.faGroup              = ""
        self.faProfile            = ""
        self.faMethod             = ""
        self.faPercentage         = ""

        # institutional (ie non-cleared) only
        self.designatedLocation = "" #used only when shortSaleSlot=2
        self.openClose     = ""     # O=Open, C=Close
        self.origin        = CUSTOMER  # 0=Customer, 1=Firm
        self.shortSaleSlot = 0         # type: int  # 1 if you hold the shares, 2 if they will be delivered from elsewhere.  Only for Action=SSHORT
        self.exemptCode    = -1

        # SMART routing only
        self.discretionaryAmt = 0
        self.optOutSmartRouting = False

        # BOX exchange orders only
        self.auctionStrategy = AUCTION_UNSET # type: int  # AUCTION_MATCH, AUCTION_IMPROVEMENT, AUCTION_TRANSPARENT
        self.startingPrice   = UNSET_DOUBLE   # type: float
        self.stockRefPrice   = UNSET_DOUBLE   # type: float
        self.delta           = UNSET_DOUBLE   # type: float

        # pegged to stock and VOL orders only
        self.stockRangeLower = UNSET_DOUBLE   # type: float
        self.stockRangeUpper = UNSET_DOUBLE   # type: float

        self.randomizePrice = False
        self.randomizeSize = False

        # VOLATILITY ORDERS ONLY
        self.volatility            = UNSET_DOUBLE  # type: float
        self.volatilityType        = UNSET_INTEGER  # type: int  # 1=daily, 2=annual
        self.deltaNeutralOrderType = ""
        self.deltaNeutralAuxPrice  = UNSET_DOUBLE  # type: float
        self.deltaNeutralConId     = 0
        self.deltaNeutralSettlingFirm = ""
        self.deltaNeutralClearingAccount = ""
        self.deltaNeutralClearingIntent = ""
        self.deltaNeutralOpenClose = ""
        self.deltaNeutralShortSale = False
        self.deltaNeutralShortSaleSlot = 0
        self.deltaNeutralDesignatedLocation = ""
        self.continuousUpdate      = False
        self.referencePriceType    = UNSET_INTEGER  # type: int  # 1=Average, 2 = BidOrAsk

        # COMBO ORDERS ONLY
        self.basisPoints     = UNSET_DOUBLE  # type: float  # EFP orders only
        self.basisPointsType = UNSET_INTEGER  # type: int  # EFP orders only

        # SCALE ORDERS ONLY
        self.scaleInitLevelSize  = UNSET_INTEGER  # type: int
        self.scaleSubsLevelSize  = UNSET_INTEGER  # type: int
        self.scalePriceIncrement = UNSET_DOUBLE  # type: float
        self.scalePriceAdjustValue = UNSET_DOUBLE  # type: float
        self.scalePriceAdjustInterval = UNSET_INTEGER  # type: int
        self.scaleProfitOffset = UNSET_DOUBLE  # type: float
        self.scaleAutoReset = False
        self.scaleInitPosition = UNSET_INTEGER   # type: int
        self.scaleInitFillQty = UNSET_INTEGER    # type: int
        self.scaleRandomPercent = False
        self.scaleTable = ""

        # HEDGE ORDERS
        self.hedgeType             = "" # 'D' - delta, 'B' - beta, 'F' - FX, 'P' - pair
        self.hedgeParam            = "" # 'beta=X' value for beta hedge, 'ratio=Y' for pair hedge

        # Clearing info
        self.account               = "" # IB account
        self.settlingFirm          = ""
        self.clearingAccount       = ""   #True beneficiary of the order
        self.clearingIntent        = "" # "" (Default), "IB", "Away", "PTA" (PostTrade)

        # ALGO ORDERS ONLY
        self.algoStrategy          = ""

        self.algoParams            = None    #TagValueList
        self.smartComboRoutingParams = None  #TagValueList

        self.algoId = ""

        # What-if
        self.whatIf = False

        # Not Held
        self.notHeld = False
        self.solicited = False

        # models
        self.modelCode = ""

        # order combo legs

        self.orderComboLegs = None  # OrderComboLegListSPtr

        self.orderMiscOptions = None  # TagValueList

        # VER PEG2BENCH fields:
        self.referenceContractId = 0
        self.peggedChangeAmount = 0.
        self.isPeggedChangeAmountDecrease = False
        self.referenceChangeAmount = 0.
        self.referenceExchangeId = ""
        self.adjustedOrderType = ""

        self.triggerPrice = UNSET_DOUBLE
        self.adjustedStopPrice = UNSET_DOUBLE
        self.adjustedStopLimitPrice = UNSET_DOUBLE
        self.adjustedTrailingAmount = UNSET_DOUBLE
        self.adjustableTrailingUnit = 0
        self.lmtPriceOffset = UNSET_DOUBLE

        self.conditions = []  # std::vector<std::shared_ptr<OrderCondition>>
        self.conditionsCancelOrder = False
        self.conditionsIgnoreRth = False

        # ext operator
        self.extOperator = ""

        # native cash quantity
        self.cashQty = UNSET_DOUBLE

        self.mifid2DecisionMaker = ""
        self.mifid2DecisionAlgo = ""
        self.mifid2ExecutionTrader = ""
        self.mifid2ExecutionAlgo = ""

        self.dontUseAutoPriceForHedge = False

        self.isOmsContainer = False

        self.discretionaryUpToLimitPrice = False

        self.autoCancelDate = ""
        self.filledQuantity = UNSET_DECIMAL
        self.refFuturesConId = 0
        self.autoCancelParent = False
        self.shareholder = ""
        self.imbalanceOnly = False
        self.routeMarketableToBbo = False
        self.parentPermId = 0

        self.usePriceMgmtAlgo = None
        self.duration = UNSET_INTEGER
        self.postToAts = UNSET_INTEGER

    def __str__(self):
        s = "%s,%d,%s:" % (self.orderId, self.clientId, self.permId)

        s += " %s %s %d@%f" % (
            self.orderType,
            self.action,
            self.totalQuantity,
            self.lmtPrice)

        s += " %s" % self.tif

        if self.orderComboLegs:
            s += " CMB("
            for leg in self.orderComboLegs:
                s += str(leg) + ","
            s += ")"

        if self.conditions:
            s += " COND("
            for cond in self.conditions:
                s += str(cond) + ","
            s += ")"

        return s



import pandas as pd
from ..event.event import Direction


class Portfolio:
    """        
    Portfolio consumes BarEvent and FillEvent to update the portfolio
    """
    def __init__(self, initial_cash, db_client=None):
        self.cash = initial_cash
        self.init_cash = initial_cash
        self.market_value = 0 #market value excluding cash
        """
        RLZD_P : Realized profit
        UNRLZD_P: Unrealized profit
        """
        self.fin_instruments = pd.DataFrame(columns=['POS', 'AVG_PRICE', 
                                                     'MKT_VAL', 'RLZD_P',
                                                     'UNRLZD_P'])
        self.db_client = db_client
    
    
    def on_bar(self, bar_event):
        """
        consumes BarEvent to update the portfolio value
        """
        if self._ticker_exist(bar_event.ticker):
            fin_instr = self.fin_instruments.loc[bar_event.ticker]
            old_market_value = fin_instr.MKT_VAL
            fin_instr.MKT_VAL = bar_event.close_price * fin_instr.POS
            fin_instr.UNRLZD_P = (bar_event.close_price - fin_instr.AVG_PRICE) \
                                 * fin_instr.POS
                                 
            self.market_value += fin_instr.MKT_VAL - old_market_value

            
    def on_fill(self, fill_event):
        #consumes FillEvent to update the portfolio
        if not self._ticker_exist(fill_event.ticker):
            #print("create a new fin instr")
            self._create_fin_instr(fill_event)      
                        
        else:
            #print("modify a new fin instr")
            self._modify_fin_instr(fill_event)


    def _create_fin_instr(self, fill_event):
        #create new entry for the ticker symbol and store it in fin_instruments
        market_value = fill_event.quantity * fill_event.price
        
        if fill_event.direction == Direction.SHORT:
            market_value *= -1
            fill_event.quantity *= -1

        fin_instr = [fill_event.quantity, fill_event.price, market_value, 0, 0]
        self.fin_instruments.loc[fill_event.ticker] = fin_instr
        self.market_value += market_value
        #self.cash -= market_value
        self.cash = self.cash - market_value - fill_event.commission


    def _modify_fin_instr(self, fill_event):
        fin_instr = self.fin_instruments.loc[fill_event.ticker]
        old_market_value = fin_instr.MKT_VAL

        if fill_event.direction == Direction.LONG:
            #add POS for a long position
            if fin_instr.POS > 0:
                total_cost = fin_instr.POS * fin_instr.AVG_PRICE + fill_event.quantity * fill_event.price 
                fin_instr.AVG_PRICE = total_cost / (fin_instr.POS + fill_event.quantity)

            #buy back a short position
            elif fin_instr.POS < 0:
                fin_instr.RLZD_P += (fin_instr.AVG_PRICE - fill_event.price) * fill_event.quantity                                   

            fin_instr.POS += fill_event.quantity                        
            #self.cash -= fill_event.quantity * fill_event.price
            self.cash = self.cash - fill_event.quantity * fill_event.price - fill_event.commission
            #self.cash -= fill_event.commission

        #fill_event.direction == Direction.SHORT:    
        else:
            #sell POS for a long position
            if fin_instr.POS > 0:       
                fin_instr.RLZD_P += (fill_event.price - fin_instr.AVG_PRICE) * fill_event.quantity                                        

            #add POS for a short position
            elif fin_instr.POS < 0:
                total_cost = -fin_instr.POS * fin_instr.AVG_PRICE + fill_event.quantity * fill_event.price                  
                fin_instr.AVG_PRICE = total_cost / (-fin_instr.POS + fill_event.quantity)

            fin_instr.POS -= fill_event.quantity  
            #self.cash += fill_event.quantity * fill_event.price
            self.cash = self.cash + fill_event.quantity * fill_event.price - fill_event.commission
            #self.cash -= fill_event.commission

        fin_instr.UNRLZD_P = (fill_event.price - fin_instr.AVG_PRICE) * fin_instr.POS
        fin_instr.MKT_VAL = fin_instr.POS * fill_event.price
        self.market_value += fin_instr.MKT_VAL - old_market_value

        #clear the instrument if all pos was sold/bought back
        if fin_instr.POS == 0:
            self.fin_instruments.drop(labels=fill_event.ticker, axis=0, inplace=True)

    
    def get_asset_val(self):
        return self.cash + self.market_value

    
    def get_fin_instrument(self, ticker):
        """
        return the whole instrucment information of specific ticker.
        return None if the ticker symbol does not exist
        """
        if not self._ticker_exist(ticker):
            return None
        
        return self.fin_instruments.loc[ticker]

    
    def _ticker_exist(self, ticker):
        if ticker in self.fin_instruments.index:
            return True   
        
        return False
        
    def save_to_db(self, time):
        if self.db_client is not None:
            query = ("INSERT INTO Portfolio (Datetime, Cash, MarketValue, AssetValue)"
                    "VALUES (%s, %s, %s, %s)"
            )
            data = (time, self.cash, float(self.market_value), float(self.get_asset_val()))
            cursor = self.db_client.cursor()
            cursor.execute(query, data)
            self.db_client.commit()         



from abc import ABC, abstractmethod
from datetime import datetime, timezone

class AbstractBroker(ABC):
    """
    The ExecutionHandler abstract class handles the interaction
    between a set of order objects generated by a PortfolioHandler
    and the set of Fill objects that actually occur in the
    market.
    
    The handlers can be used to subclass simulated brokerages
    or live brokerages, with identical interfaces. This allows
    strategies to be backtested in a very similar manner to the
    live trading engine.
    """    
    @abstractmethod
    def execute_order(self):
        pass
    
    
    def save_to_db(self, fill_event):
        query = ("INSERT INTO FillEvent (Datetime, Ticker, Direction, Exchange, Price, Quantity, Commission)"
                "VALUES (%s, %s, %s, %s, %s, %s, %s)"
        )
        timestamp_converted = datetime.fromtimestamp(fill_event.timestamp, timezone.utc)
        data = (timestamp_converted, fill_event.ticker, fill_event.direction.name, 
                fill_event.exchange, fill_event.price, fill_event.quantity, fill_event.commission)
        cursor = self.db_client.cursor()
        cursor.execute(query, data)
        self.db_client.commit()   



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Nov 28 12:51:56 2021

@author: ron
"""
import unittest
from tradingsystem.order_handler.fixed_order_handler import FixedOrderHandler
from tradingsystem.order_handler.max_order_handler import MaxOrderHandler
from tradingsystem.event.event import SignalEvent, OrderType, Direction
from tradingsystem.portfolio.portfolio import Portfolio
from datetime import datetime
from tradingsystem.common import SessionType
import queue


class TestPositionHandler(unittest.TestCase):   
    def setUp(self):        
        self.portfolio = Portfolio(10000)
        self.date_time = datetime(2021, 11, 5, 15, 30, 20)
        self.event_queue = queue.Queue()

      
    def test_fixed_position(self):
        fixed = FixedOrderHandler(5, self.portfolio, self.event_queue, SessionType.BACKTEST)
        #test buy
        signal = SignalEvent(self.date_time, 'GOOG', OrderType.MARKET, Direction.LONG, 2821.990)
        fixed.create_order(signal)
        order_event = self.event_queue.get()
        self.assertEqual(order_event.quantity, 5)
        #test sell
        signal = SignalEvent(self.date_time, 'GOOG', OrderType.MARKET, Direction.SHORT, 2821.990)
        fixed.create_order(signal)
        order_event = self.event_queue.get()
        self.assertEqual(order_event.quantity, 5)
        
        
    def test_max_position(self):
        max_order = MaxOrderHandler(self.portfolio, self.event_queue, SessionType.BACKTEST)
        #test buy
        signal = SignalEvent(self.date_time, 'GOOG', OrderType.MARKET, Direction.LONG, 2821.990)
        max_order.create_order(signal)
        order_event = self.event_queue.get()
        self.assertEqual(order_event.quantity, 3)
        #test sell
        signal = SignalEvent(self.date_time, 'GOOG', OrderType.MARKET, Direction.SHORT, 1230)
        max_order.create_order(signal)
        order_event = self.event_queue.get()
        self.assertEqual(order_event.quantity, 8)

    
    def test_max_position_not_enough_cash(self):
        max_order = MaxOrderHandler(self.portfolio, self.event_queue, SessionType.BACKTEST)
        #test buy
        signal = SignalEvent(self.date_time, 'GOOG', OrderType.MARKET, Direction.LONG, 50000)
        max_order.create_order(signal)
        self.assertEqual(self.event_queue.qsize(), 0)





if __name__ == '__main__':
    unittest.main(verbosity=2)     
        
        


from datetime import datetime, time, timezone, timedelta, date
from .base import BaseSchedule


class FXSchedule(BaseSchedule):
    def __init__(self, year=datetime.utcnow().year):
        super().__init__()
        self._gen_weekends(year)
        

    def get_trading_hours(self, date):
        """
        TODO if the current time is between the trading time, start immediately
            rather than wait for the next open day
        """
        # if self.calendar.is_holiday(date):
        #     start = datetime.combine(date, time(21, 00, tzinfo=timezone.utc)).timestamp()
        # else:
        #     start = ti.time()
        start = datetime.combine(date, time(21, 15, tzinfo=timezone.utc)).timestamp()
        end = datetime.combine(date + timedelta(days=1), time(21, 00, tzinfo=timezone.utc)).timestamp()      
        return (start, end)
    
    
    def _gen_weekends(self, year):
        d = date(year, 1, 1)
        while d != date(year + 1, 1, 1):
            if 5 <= d.isoweekday() <= 6:
                self.calendar.add_holiday(d)
            d += timedelta(days=1)


    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Nov 14 11:36:51 2021

@author: ron
"""
from .base import AbstractTechnicalIndicator
from .sma import SMA
import numpy as np


class BBData:
    def __init_(self):
        self.upper = 0
        self.mid = 0
        self.lower = 0


class BB(AbstractTechnicalIndicator):
    
    def __init__(self, period, multiple):  
        """
        Parameters
        ----------
        data_store : list of lists
                   First list stores upperbound value
                   Second list stores mid value
                   Third list stores lowerbound value
        """
        self.sma = SMA(period)
        self.data_store = [[], [], []]
        self.lastest_val = BBData()
        self.k = multiple
        self.ready = False


    def update(self, price):
        #calculate lastest sma
        self.sma.update(price)
        if not self.ready and self.sma:
            self.ready = True
              
            
        if self.ready:
            mid = self.sma.lastest_val
            sd = np.std(self.sma.data)
            upper = mid + self.k * sd
            lower = mid - self.k * sd
            self.lastest_val.upper = upper
            self.lastest_val.mid = mid
            self.lastest_val.lower = lower
            self.data_store[0].append(upper)
            self.data_store[1].append(mid)
            self.data_store[2].append(lower)


    def __bool__(self):
        return self.ready

    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec 18 19:32:03 2021

@author: ron
"""
import queue
from .event.event import EventType
import time
from datetime import datetime, timezone, timedelta



class TradingSession:
    """
    Enscapsulates the settings and components for
    carrying out a live trading session through Interactive Brokers.
    """    
    def __init__(
        self, twsclient, price_handler, 
        strategy, portfolio, 
        order_handler, broker, events_queue,
        trading_schedule, trading_end, statistics=None, 
        ):
        """
        TODO
        Now it only supports UTC time, it should support other time zones
        """
        self.strategy = strategy
        self.events_queue = events_queue
        self.price_handler = price_handler
        self.portfolio = portfolio
        self.order_handler = order_handler
        self.broker = broker
        self.statistics = statistics
        self.twsclient = twsclient
        self.previous_day = None
        self.trading_end = trading_end.replace(tzinfo=timezone.utc).timestamp()
        self.trading_schedule = trading_schedule

             
    def _run_session(self):
        """
        Carries out an infinite while loop that polls the
        events queue and directs each event to either the
        strategy component of the execution handler. The
        loop continue until the event queue has been
        emptied.
        """
        if self.trading_end < time.time():
            raise ValueError("invalid trading period, trading_end is less than current time!")
        
        while time.time() <= self.trading_end:
            today = datetime.utcnow().date()      
            if not self.trading_schedule.calendar.is_holiday(today):
                start, end = self.trading_schedule.get_trading_hours(today)
                print("start: ", start)
                print("end: ", end)
                cur_time = time.time()
                if cur_time < start:
                    sleep_time = start - cur_time
                    print(f"sleep {sleep_time} to trade...")
                    time.sleep(sleep_time)
                self.twsclient.end_time = end
                self.twsclient.connect()                
                self.twsclient.run()
                self.price_handler.request_data()
                self._event_loop(start, end)  
                                        
            self._reset()       
            self._sleep_next_open_day()


    def _event_loop(self, start, end):
        """
        end + 10: give some time to get the last data
        """
        while start <= time.time() <= end + 10:
            self.price_handler.get_next()
            self.price_handler.check_bar_interruption()
            self.broker.get_fill_event()
            try:                
                event = self.events_queue.get(False)
            except queue.Empty:
                pass            
            else:
                day = datetime.fromtimestamp(event.timestamp, timezone.utc).date()
                if self.previous_day is None:
                    self.previous_day = day
                elif day > self.previous_day:
                    self.statistics.update(self.previous_day, self.portfolio.get_asset_val())
                    self.portfolio.save_to_db(self.previous_day)
                    self.previous_day = day

                if event is not None:
                    if event.type == EventType.BAR:
                        self.strategy.calculate_signal(event)
                        self.portfolio.on_bar(event)                 
                    elif event.type == EventType.SIGNAL:
                        self.order_handler.create_order(event)
                    elif event.type == EventType.ORDER:
                        self.broker.execute_order(event)
                    elif event.type == EventType.FILL:
                        self.portfolio.on_fill(event)
                    else:
                        raise NotImplementedError("Unsupported event.type '%s'" % event.type)
     
                                      
    def _reset(self):
        self.price_handler.cancel_subscription()
        if self.twsclient.isConnected():
            print("disconnect tws...")
            self.twsclient.disconnect()        
         
         
    def _sleep_next_open_day(self):
        day = datetime.utcnow().date()
        while self.trading_schedule.calendar.is_holiday(day):
            day += timedelta(days=1)
        start, _ = self.trading_schedule.get_trading_hours(day)
        cur_time = time.time()
        sleep_time = start - cur_time
        print(f"sleep {sleep_time} to trade...")
        time.sleep(max(0, sleep_time))
        
        
    def start_trading(self, testing=False):
        """
        Runs either a backtest or live session, and outputs performance when complete.
        """
        self._run_session()
        print("disconnecting tws...")
        self.twsclient.disconnect()
        print("---------------------------------")
        print('Ending cash: ' + str(self.portfolio.cash))
        print('Ending market value: ' + str(self.portfolio.market_value))
        print('Ending asset value: ' + str(self.portfolio.get_asset_val()))
        print('Ending Portfolio:')
        results = self.statistics.get_results()
        print("---------------------------------")
        print("Strategy:", self.strategy)
        print("Sharpe Ratio: %0.2f" % results["sharpe_ratio"])
        print(
            "Max Drawdown: %0.2f%%" % (
                results["max_drawdown"] * 100.0
            )
        )
        print("Return: %0.2f%%" % results["return"])
        print("Trading ends")
        self.statistics.plot_results()







"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


"""
Collection of misc tools
"""


import sys
import logging
import inspect

from decimal import Decimal
from .common import UNSET_INTEGER, UNSET_DOUBLE, UNSET_LONG, UNSET_DECIMAL


logger = logging.getLogger(__name__)


# I use this just to visually emphasize it's a wrapper overridden method
def iswrapper(fn):
    return fn


class BadMessage(Exception):
    def __init__(self, text):
        self.text = text

class ClientException(Exception):
    def __init__(self, code, msg, text):
        self.code = code
        self.msg = msg
        self.text = text

class LogFunction(object):
    def __init__(self, text, logLevel):
        self.text = text
        self.logLevel = logLevel

    def __call__(self, fn):
        def newFn(origSelf, *args, **kwargs):
            if logger.getLogger().isEnabledFor(self.logLevel):
                argNames = [argName for argName in inspect.getfullargspec(fn)[0] if argName != 'self']
                logger.log(self.logLevel,
                    "{} {} {} kw:{}".format(self.text, fn.__name__,
                        [nameNarg for nameNarg in zip(argNames, args) if nameNarg[1] is not origSelf], kwargs))
            fn(origSelf, *args)
        return newFn


def current_fn_name(parent_idx = 0):
    #depth is 1 bc this is already a fn, so we need the caller
    return sys._getframe(1 + parent_idx).f_code.co_name


def setattr_log(self, var_name, var_value):
    #import code; code.interact(local=locals())
    logger.debug("%s %s %s=|%s|", self.__class__, id(self), var_name, var_value)
    super(self.__class__, self).__setattr__(var_name, var_value)


SHOW_UNSET = True


def decode(the_type, fields, show_unset = False):
    try:
        s = next(fields)
    except StopIteration:
        raise BadMessage("no more fields")

    logger.debug("decode %s %s", the_type, s)
    
    if the_type is Decimal:
        if s is None or len(s) == 0 or s.decode() == "2147483647" or s.decode() == "9223372036854775807" or s.decode() == "1.7976931348623157E308":
            return UNSET_DECIMAL
        else:
            return the_type(s.decode())

    if the_type is str:
        if type(s) is str:
            return s
        elif type(s) is bytes:
            return s.decode(errors='backslashreplace')
        else:
            raise TypeError("unsupported incoming type " + type(s) + " for desired type 'str")

    orig_type = the_type
    if the_type is bool:
        the_type = int

    if show_unset:
        if s is None or len(s) == 0:
            if the_type is float:
                n = UNSET_DOUBLE
            elif the_type is int:
                n = UNSET_INTEGER
            else:
                raise TypeError("unsupported desired type for empty value" + the_type)
        else:
            n = the_type(s)
    else:
        n = the_type(s or 0)

    if orig_type is bool:
        n = False if n == 0 else True

    return n


def ExerciseStaticMethods(klass):

    import types
    #import code; code.interact(local=dict(globals(), **locals()))
    for (_, var) in inspect.getmembers(klass):
        #print(name, var, type(var))
        if type(var) == types.FunctionType:
            print("Exercising: %s:" % var)
            print(var())
            print()
            
           
def floatToStr(val):
    return str(val) if val != UNSET_DOUBLE else ""

 
def longToStr(val):
    return str(val) if val != UNSET_LONG else ""

def isAsciiPrintable(val):
    return all(ord(c) >=32 and ord(c) < 128 for c in val)

def decimalMaxString(val: Decimal):
    return "{:f}".format(val) if val != UNSET_DECIMAL else ""
